searchData={"items":[{"type":"behaviour","title":"Broadway","doc":"Broadway is a concurrent, multi-stage tool for building\ndata ingestion and data processing pipelines.\n\nIt allows developers to consume data efficiently from different\nsources, such as Amazon SQS, Apache Kafka, Google Cloud PubSub,\nRabbitMQ and others.","ref":"Broadway.html"},{"type":"behaviour","title":"Built-in features - Broadway","doc":"* Back-pressure - by relying on `GenStage`, we only get the amount\n    of events necessary from upstream sources, never flooding the\n    pipeline.\n\n  * Automatic acknowledgements - Broadway automatically acknowledges\n    messages at the end of the pipeline or in case of errors.\n\n  * Batching - Broadway provides built-in batching, allowing you to\n    group messages either by size and/or by time. This is important\n    in systems such as Amazon SQS, where batching is the most efficient\n    way to consume messages, both in terms of time and cost.\n\n  * Fault tolerance - Broadway pipelines are carefully designed to manage\n    failures. Producers are isolated from the rest of the pipeline and\n    automatically resubscribe to your data source in case of crashes.\n    At the same time, all of your Broadway callbacks are stateless, which\n    allows Broadway to handle any errors locally. This provides a stable\n    foundation that play well with your producers, regardless if their\n    delivery guarantees are at least once, at most once, or exactly once.\n\n  * Graceful shutdown - Broadway integrates with the VM to provide graceful\n    shutdown. By starting Broadway as part of your supervision tree, it will\n    guarantee all events are flushed once the VM shuts down.\n\n  * Built-in testing - Broadway ships with a built-in test API, making it\n    easy to push test messages through the pipeline and making sure the\n    event was properly processed.\n\n  * Custom failure handling - Broadway provides a `c:handle_failed/2` callback\n    where developers can outline custom code to handle errors. For example,\n    if they want to move messages to another queue for further processing.\n\n  * Dynamic batching - Broadway allows developers to batch messages based on\n    custom criteria. For example, if your pipeline needs to build\n    batches based on the `user_id`, email address, etc, it can be done\n    by calling `Broadway.Message.put_batch_key/2`.\n\n  * Ordering and Partitioning - Broadway allows developers to partition\n    messages across workers, guaranteeing messages within the same partition\n    are processed in order. For example, if you want to guarantee all events\n    tied to a given `user_id` are processed in order and not concurrently,\n    you can set the `:partition_by` option. See [\"Ordering and partitioning\"](#module-ordering-and-partitioning).\n\n  * Rate limiting - Broadway allows developers to rate limit all producers in\n    a single node by a given number of messages in a time period, allowing\n    developers to easily work sources or sinks that cannot cope with a high\n    number of requests. See the `:rate_limiting` option for producers in\n    `start_link/2`.\n\n  * Metrics - Broadway uses the `:telemetry` library for instrumentation,\n    see [\"Telemetry\"](#module-telemetry) section below for more information.","ref":"Broadway.html#module-built-in-features"},{"type":"behaviour","title":"The Broadway behaviour - Broadway","doc":"In order to use Broadway, you need to:\n\n  1. Define your pipeline configuration\n  2. Define a module implementing the Broadway behaviour","ref":"Broadway.html#module-the-broadway-behaviour"},{"type":"behaviour","title":"Example - Broadway","doc":"Broadway is a process-based behaviour, and you begin by\ndefining a module that invokes `use Broadway`. Processes\ndefined by these modules will often be started by a\nsupervisor, and so a `start_link/1` function is frequently\nalso defined but not strictly necessary.\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(MyBroadway,\n          name: MyBroadwayExample,\n          producer: [\n            module: {Counter, []},\n            concurrency: 1\n          ],\n          processors: [\n            default: [concurrency: 2]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nThen add your Broadway pipeline to your supervision tree\n(usually in `lib/my_app/application.ex`):\n\n    children = [\n      {MyBroadway, []}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n\nAdding your pipeline to your supervision tree in this way\ncalls the default `child_spec/1` function that is generated\nwhen `use Broadway` is invoked. If you would like to customize\nthe child spec passed to the supervisor, you can override the\n`child_spec/1` function in your module or explicitly pass a\nchild spec to the supervisor when adding it to your supervision tree.\n\nThe configuration above defines a pipeline with:\n\n  * One producer\n  * Two processors\n\nHere is how this pipeline would be represented:\n\n```asciidoc\n                     [producer_1]\n                         / \\\n                        /   \\\n                       /     \\\n                      /       \\\n             [processor_1] [processor_2]   <- process each message\n```\n\nAfter the pipeline is defined, you need to implement the `c:handle_message/3`\ncallback which will be invoked by processors for each message.\n\n`c:handle_message/3` receives every message as a `Broadway.Message`\nstruct and it must return an updated message.","ref":"Broadway.html#module-example"},{"type":"behaviour","title":"Batching - Broadway","doc":"Depending on the scenario, you may want to group processed messages as\nbatches before publishing your data. This is common and especially\nimportant when working with services like AWS S3 and SQS that provide a\nspecific API for sending and retrieving batches. This can drastically\nincrease throughput and consequently improve the overall performance of\nyour pipeline.\n\nTo create batches, define the `:batchers` configuration option:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(MyBroadway,\n          name: MyBroadwayExample,\n          producer: [\n            module: {Counter, []},\n            concurrency: 1\n          ],\n          processors: [\n            default: [concurrency: 2]\n          ],\n          batchers: [\n            sqs: [concurrency: 2, batch_size: 10],\n            s3: [concurrency: 1, batch_size: 10]\n          ]\n        )\n      end\n\n      # ...callbacks...\n    end\n\nThe configuration above defines a pipeline with:\n\n  * One producer\n  * Two processors\n  * One batcher named `:sqs` with two batch processors\n  * One batcher named `:s3` with one batch processor\n\nHere is how this pipeline would be represented:\n\n```asciidoc\n                     [producer_1]\n                         / \\\n                        /   \\\n                       /     \\\n                      /       \\\n             [processor_1] [processor_2]   <- process each message\n                      /\\     /\\\n                     /  \\   /  \\\n                    /    \\ /    \\\n                   /      x      \\\n                  /      / \\      \\\n                 /      /   \\      \\\n                /      /     \\      \\\n           [batcher_sqs]    [batcher_s3]\n                /\\                  \\\n               /  \\                  \\\n              /    \\                  \\\n             /      \\                  \\\n   [batch_sqs_1] [batch_sqs_2]    [batch_s3_1] <- process each batch\n```\n\nAdditionally, you have to define the `c:handle_batch/4` callback,\nwhich batch processors invoke for each batch. You can then\ncall `Broadway.Message.put_batcher/2` inside `c:handle_message/3` to\ncontrol which batcher the message should go to.\n\nThe batcher receives processed messages and creates batches\nspecified by the `batch_size` and `batch_timeout` configuration. The\ngoal is to create a batch with at most `batch_size` entries within\n`batch_timeout` milliseconds. Each message goes into a particular batch,\ncontrolled by calling `Broadway.Message.put_batch_key/2` in\n`c:handle_message/3`. Once a batch is created in the batcher, it is sent\nto a separate process (the batch processor) that will call `c:handle_batch/4`,\npassing the batcher, the batch itself (a list of messages), a `Broadway.BatchInfo`\nstruct, and the Broadway context.\n\nFor example, imagine your producer generates integers as `data`.\nYou want to route the odd integers to SQS and the even ones to\nS3. Your pipeline would look like this:\n\n    defmodule MyBroadway do\n      use Broadway\n      import Integer\n\n      alias Broadway.Message\n\n      # ...start_link...\n\n      @impl true\n      def handle_message(_, %Message{data: data} = message, _) when is_odd(data) do\n        message\n        |> Message.update_data(&process_data/1)\n        |> Message.put_batcher(:sqs)\n      end\n\n      def handle_message(_, %Message{data: data} = message, _) when is_even(data) do\n        message\n        |> Message.update_data(&process_data/1)\n        |> Message.put_batcher(:s3)\n      end\n\n      defp process_data(data) do\n        # Do some calculations, generate a JSON representation, etc.\n      end\n\n      @impl true\n      def handle_batch(:sqs, messages, _batch_info, _context) do\n        # Send batch of successful messages as ACKs to SQS\n        # This tells SQS that this list of messages were successfully processed\n      end\n\n      def handle_batch(:s3, messages, _batch_info, _context) do\n        # Send batch of messages to S3\n      end\n    end\n\nSee the [callbacks documentation](#callbacks) for more information on the\narguments given to each callback and their expected return types.","ref":"Broadway.html#module-batching"},{"type":"behaviour","title":"The default batcher - Broadway","doc":"Once you define the `:batchers` configuration key for your Broadway pipeline,\nthen **all messages get batched**. By default, unless you call\n`Broadway.Message.put_batcher/2`, messages have their batcher set to the\n`:default` batcher. If you don't define configuration for it, Broadway is going\nto raise an error.\n\nFor example, imagine you want to batch \"special\" messages and handle them differently\nthen all other messages. You can configure your pipeline like this:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(MyBroadway,\n          name: MyBroadwayExample,\n          producer: [\n            module: {Counter, []},\n            concurrency: 1\n          ],\n          processors: [\n            default: [concurrency: 2]\n          ],\n          batchers: [\n            special: [concurrency: 2, batch_size: 10],\n            default: [concurrency: 1, batch_size: 10]\n          ]\n        )\n      end\n\n      def handle_message(_, message, _) do\n        if special?(message) do\n          message\n          |> Broadway.Message.put_batcher(:special)\n        else\n          message\n        end\n      end\n\n      def handle_batch(:special, messages, _batch_info, _context) do\n        # Handle special batch\n      end\n\n      def handle_batch(:default, messages, _batch_info, _context) do\n        # Handle all other messages in batches\n      end\n\nNow you are ready to get started. See the `start_link/2` function\nfor a complete reference on the arguments and options allowed.\n\nAlso makes sure to check out GUIDES in the documentation sidebar\nfor more examples, how tos and more.","ref":"Broadway.html#module-the-default-batcher"},{"type":"behaviour","title":"Acknowledgements and failures - Broadway","doc":"At the end of the pipeline, messages are automatically acknowledged.\n\nIf there are no batchers, the acknowledgement will be done by processors.\nThe number of messages acknowledged, assuming the pipeline is running\nat full scale, will be `max_demand - min_demand`. Since the default values\nare 10 and 5 respectively, we will be acknowledging in groups of 5.\n\nIf there are batchers, the acknowledgement is done by the batchers,\nusing the `batch_size`.\n\nIn case of failures, Broadway does its best to keep the failures\ncontained and avoid losing messages. The failed message or batch is\nacknowledged as failed immediately. For every failure, a log report\nis also emitted. If your Broadway module also defines the\n`c:handle_failed/2` callback, that callback will be invoked with\nall the failed messages before they get acknowledged.\n\nNote however, that `Broadway` does not provide any sort of retries\nout of the box. This is left completely as a responsibility of the\nproducer. For instance, if you are using Amazon SQS, the default\nbehaviour is to retry unacknowledged messages after a user-defined\ntimeout. If you don't want unacknowledged messages to be retried,\nis your responsibility to configure a dead-letter queue as target\nfor those messages.","ref":"Broadway.html#module-acknowledgements-and-failures"},{"type":"behaviour","title":"Producer concurrency - Broadway","doc":"Setting producer concurrency is a tradeoff between latency and internal\nqueueing.\n\nFor efficiency, you should generally limit the amount of internal queueing.\nWhenever additional messages are sitting in a busy processor's mailbox, they\ncan't be delivered to another processor which may be available or become\navailable first.\n\nOne possible cause of internal queueing is multiple producers. This is because\neach processor's demand will be sent to all producers. For example, if a\nprocessor demands `2` messages and there are `2` producers, each producer\nwill try to produce `2` messages (for example, by pulling from a queue or\nwhatever the specific producer does) and give them to the processor. So the\nprocessor may receive `max_demand *  ` messages.\n\nSetting producer `concurrency: 1` will reduce internal queueing. This is\nlikely a good choice for producers which take minimal time to produce a\nmessage, such as `BroadwayRabbitMQ`, which receives messages as they are\npushed by RabbitMQ and can specify how many to prefetch.\n\nOn the other hand, when using a producer such as `BroadwaySQS` which must\nmake a network round trip to fetch from an external source, it may be better\nto use multiple producers and accept some internal queueing to avoid having\nfetch messages whenever there is new demand.\n\nMeasure your system to decide which setting is most appropriate.\n\nAdding another single-producer pipeline, or another node running the\npipeline, are other ways you may consider to increase throughput.","ref":"Broadway.html#module-producer-concurrency"},{"type":"behaviour","title":"Batcher concurrency - Broadway","doc":"If a batcher's `concurrency` is greater than `1`, Broadway will use as few of\nthe batcher processes as possible at any given moment, attempting to satisfy\nthe `batch_size` of one batcher process within the `batch_timeout` before\nsending messages to another.","ref":"Broadway.html#module-batcher-concurrency"},{"type":"behaviour","title":"Testing - Broadway","doc":"Many producers receive data from external systems and hitting the network\nis usually undesirable when running the tests.\n\nFor testing purposes, we recommend developers to use `Broadway.DummyProducer`.\nThis producer does not produce any messages by itself and instead the\n`test_message/3` and `test_batch/3` functions should be used to publish\nmessages.\n\nWith `test_message/3`, you can push a message into the pipeline and receive\na process message when the pipeline acknowledges the data you have pushed\nhas been processed.\n\nLet's see an example. Imagine the following `Broadway` module:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link() do\n        producer_module = Application.fetch_env!(:my_app, :producer_module)\n        producer_options = Application.get_env(:my_app, :producer_options, [])\n\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module: {producer_module, producer_options}\n          ],\n          processors: [\n            default: []\n          ],\n          batchers: [\n            default: [batch_size: 10]\n          ]\n        )\n      end\n\n      @impl true\n      def handle_message(_processor, message, _context) do\n        message\n      end\n\n      @impl true\n      def handle_batch(_batcher, messages, _batch_info, _context) do\n        messages\n      end\n    end\n\nNow in config/test.exs you could do:\n\n    config :my_app,\n      producer_module: Broadway.DummyProducer,\n      producer_options: [] # change if required for your dev/prod producer\n\nAnd we can test it like this:\n\n    defmodule MyBroadwayTest do\n      use ExUnit.Case, async: true\n\n      test \"test message\" do\n        ref = Broadway.test_message(MyBroadway, 1)\n        assert_receive {:ack, ^ref, [%{data: 1}], []}\n      end\n    end\n\nNote that at the end we received a message in the format of:\n\n    {:ack, ^ref, successful_messages, failure_messages}\n\nYou can use the acknowledgment to guarantee the message has been\nprocessed and therefore any side-effect from the pipeline should be\nvisible.\n\nWhen using `test_message/3`, the message will be delivered as soon as\npossible, without waiting for the pipeline `batch_size` to be reached\nor without waiting for `batch_timeout`. This behaviour is useful to test\nand verify single messages, without imposing high timeouts to our test\nsuites.\n\nIn case you want to test multiple messages, then you need to use\n`test_batch/3`. `test_batch/3` will respect the batching configuration,\nwhich most likely means you need to increase your test timeouts:\n\n    test \"batch messages\" do\n      ref = Broadway.test_batch(MyBroadway, [1, 2, 3])\n      assert_receive {:ack, ^ref, [%{data: 1}, %{data: 2}, %{data: 3}], []}, 1000\n    end\n\nHowever, keep in mind that, generally speaking, there is no guarantee\nthe messages will arrive in the same order that you have sent them,\nespecially for large batches, as Broadway will process large batches\nconcurrently and order will be lost.\n\nIf you want to send more than one test message at once, then we recommend\nsetting the `:batch_mode` to `:bulk`, especially if you want to assert how\nthe code will behave with large batches. Otherwise the batcher will flush\nmessages as soon as possible and in small batches.\n\nHowever, keep in mind that, regardless of the `:batch_mode` you cannot\nrely on ordering, as Broadway pipelines are inherently concurrent. For\nexample, if you send those messages:\n\n    test \"multiple batch messages\" do\n      ref = Broadway.test_batch(MyBroadway, [1, 2, 3, 4, 5, 6, 7], batch_mode: :bulk)\n      assert_receive {:ack, ^ref, [%{data: 1}], []}, 1000\n    end","ref":"Broadway.html#module-testing"},{"type":"behaviour","title":"Testing with Ecto - Broadway","doc":"If you are using Ecto in your Broadway processors and you want\nto run your tests concurrently, you need to tell Broadway to\nuse the Ecto SQL Sandbox during tests. This can be done in two\nsteps.\n\nFirst, when you call `test_messages/3` in your tests, include\nthe `:ecto_sandbox` process in the message metadata:\n\n    Broadway.test_message(MyApp.Pipeline, message, metadata: %{ecto_sandbox: self()})\n\nNow we can use Broadway telemetry callbacks to fetch the sandbox\nprocess and enable it inside the processor. Add to your\n`test/test_helper.exs`:\n\n    defmodule BroadwayEctoSandbox do\n      def attach(repo) do\n        events = [\n          [:broadway, :processor, :start],\n          [:broadway, :batch_processor, :start],\n        ]\n\n        :telemetry.attach_many({__MODULE__, repo}, events, &__MODULE__.handle_event/4, %{repo: repo})\n      end\n\n      def handle_event(_event_name, _event_measurement, %{messages: messages}, %{repo: repo}) do\n        with [%Broadway.Message{metadata: %{ecto_sandbox: pid}} | _] <- messages do\n          Ecto.Adapters.SQL.Sandbox.allow(repo, pid, self())\n        end\n\n        :ok\n      end\n    end\n\n    BroadwayEctoSandbox.attach(MyApp.Repo)\n\nAnd now you should have concurrent Broadway tests that talk to the database.","ref":"Broadway.html#module-testing-with-ecto"},{"type":"behaviour","title":"Ordering and partitioning - Broadway","doc":"By default, Broadway processes all messages and batches concurrently,\nwhich means ordering is not guaranteed. Some producers may impose some\nordering (for instance, Apache Kafka), but if the ordering comes from a\nbusiness requirement, you will have to impose the ordering yourself.\nThis can be done with the `:partition_by` option, which enforces that\nmessages with a given property are always forwarded to the same stage.\n\nIn order to provide partitioning throughout the whole pipeline, just\nset `:partition_by` at the root of your configuration:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module: {Counter, []},\n            concurrency: 1\n          ],\n          processors: [\n            default: [concurrency: 2]\n          ],\n          batchers: [\n            sqs: [concurrency: 2, batch_size: 10],\n            s3: [concurrency: 1, batch_size: 10]\n          ],\n          partition_by: &partition/1\n        )\n      end\n\n      defp partition(msg) do\n        msg.data.user_id\n      end\n\nIn the example above, we are partitioning the pipeline by `user_id`.\nThis means any message with the same `user_id` will be handled by\nthe same processor and batch processor.\n\nThe `partition` function must return a non-negative integer,\nstarting at zero, which is routed to a stage by using the `remainder`\noption.\n\nIf the data you want to partition by is not an integer, you can\nexplicitly hash it by calling `:erlang.phash2/1`. However, note\nthat `hash` does not guarantee an equal distribution of events\nacross partitions. So some partitions may be more overloaded than\nothers, slowing down the whole pipeline.\n\nIn the example above, we have set the same partition for all\nprocessors and batchers. You can also specify the `:partition_by`\nfunction for each \"processor\" and \"batcher\" individually.\n\n> #### Even partitions {: .warning}\n>\n> Broadway partitions assume an even distribution of partitions.\n> This means that, if one partition is slow, it will slow down\n> all order partitions. This implies two things:\n>\n> * Using `:partition_by` with a high level of concurrency can\n>   actually be detrimental to performance. For example, if\n>   concurrency is set to 100, you need all 100 processors to\n>   make progress at the same time.\n>\n> * Avoid using `:partition_by` with a low value of `min_demand`.\n>   For example, setting `max_demand` to 1 (which implies `min_demand`\n>   of 0), means that each processor will receive a single message\n>   and only receive further messages once all processors complete.\n>\n> When partitioning, the default values for concurrency (which is\n> equal to the number of cores) and max_demand (which is equal to\n> 10), are good starting points.\n\n> #### Error semantics {: .warning}\n>\n> Beware of the error semantics when using partitioning.\n> If you require messages to be processed in order and a message\n> fails, the partition will continue processing messages, which\n> may be undesired. If your producer supports retrying, the\n> failed message may be retried later, out of its original order.\n> Those issues happen regardless of Broadway and solutions to said\n> problems almost always need to be addressed outside of Broadway too.","ref":"Broadway.html#module-ordering-and-partitioning"},{"type":"behaviour","title":"Configuration storage - Broadway","doc":"Broadway stores configuration globally in a configurable storage method.\nBroadway comes with two configuration storage options:\n\n  * `:persistent_term` (the default)\n  * `:ets`","ref":"Broadway.html#module-configuration-storage"},{"type":"behaviour","title":"Persistent term - Broadway","doc":"This is the most efficient option for static Broadway pipeline definitions,\nas this option never deletes the Broadway configuration from storage. It's based\non [`:persistent_term`](`:persistent_term`).\n\n```elixir\nconfig :broadway, config_storage: :persistent_term\n```\n\nThe speed of storing and updating using `:persistent_term` is proportional\nto the number of already-created terms in the storage. If you are creating\nseveral Broadway pipelines dynamically, that may affect the persistent term\nstorage performance. Furthermore, even if you are restarting the same pipeline\nbut you are using different parameters each time, that will require a global\ngarbage collection pass to update the `:persistent_term` configuration.\nIf you are starting Broadway pipelines dynamically, you must use the `:ets`\nstorage.","ref":"Broadway.html#module-persistent-term"},{"type":"behaviour","title":"ETS - Broadway","doc":"An [ETS](`:ets`)-backed configuration storage, useful if Broadway pipelines are\nstarted dynamically. To use this configuration storage option, configure the\n`:broadway` application in :your configuration\n\n```elixir\n# For example, in config/config.exs:\nconfig :broadway, config_storage: :ets\n```\n\nUsing `:ets` as the config storage will allow for a dynamic number of Broadway server\nconfigurations to be stored and fetched without the associated performance trade-offs\nthat `:persistent_term` has.","ref":"Broadway.html#module-ets"},{"type":"behaviour","title":"Telemetry - Broadway","doc":"Broadway currently exposes following Telemetry events:\n\n  * `[:broadway, :topology, :init]` - Dispatched when the topology for\n    a Broadway pipeline is initialized. The config key in the metadata\n    contains the configuration options that were provided to\n    `Broadway.start_link/2`.\n\n    * Measurement: `%{system_time: integer}`\n    * Metadata: `%{supervisor_pid: pid, config: keyword}`\n\n  * `[:broadway, :processor, :start]` - Dispatched by a Broadway processor\n    before the optional `c:prepare_messages/2`\n\n    * Measurement: `%{system_time: integer}`\n    * Metadata:\n\n      ```\n      %{\n        topology_name: atom,\n        name: atom,\n        processor_key: atom,\n        index: non_neg_integer,\n        messages: [Broadway.Message.t],\n        telemetry_span_context: reference,\n        producer: {atom, list}\n      }\n      ```\n\n  * `[:broadway, :processor, :stop]` -  Dispatched by a Broadway processor\n    after `c:prepare_messages/2` and after all `c:handle_message/3` callback\n    has been invoked for all individual messages\n\n    * Measurement: `%{duration: native_time}`\n\n    * Metadata:\n\n      ```\n      %{\n        topology_name: atom,\n        name: atom,\n        processor_key: atom,\n        index: non_neg_integer,\n        successful_messages_to_ack: [Broadway.Message.t],\n        successful_messages_to_forward: [Broadway.Message.t],\n        failed_messages: [Broadway.Message.t],\n        telemetry_span_context: reference,\n        producer: {atom, list}\n      }\n      ```\n\n  * `[:broadway, :processor, :message, :start]` - Dispatched by a Broadway processor\n    before your `c:handle_message/3` callback is invoked\n\n    * Measurement: `%{system_time: integer}`\n\n    * Metadata:\n\n      ```\n      %{\n        processor_key: atom,\n        topology_name: atom,\n        name: atom,\n        index: non_neg_integer,\n        message: Broadway.Message.t,\n        telemetry_span_context: reference\n      }\n      ```\n\n  * `[:broadway, :processor, :message, :stop]` - Dispatched by a Broadway processor\n    after your `c:handle_message/3` callback has returned\n\n    * Measurement: `%{duration: native_time}`\n\n    * Metadata:\n\n      ```\n      %{\n        processor_key: atom,\n        topology_name: atom,\n        name: atom,\n        index: non_neg_integer,\n        message: Broadway.Message.t,\n        telemetry_span_context: reference\n      }\n      ```\n\n  * `[:broadway, :processor, :message, :exception]` - Dispatched by a Broadway processor\n    if your `c:handle_message/3` callback encounters an exception\n\n    * Measurement: `%{duration: native_time}`\n\n    * Metadata:\n\n      ```\n      %{\n        processor_key: atom,\n        topology_name: atom,\n        name: atom,\n        index: non_neg_integer,\n        message: Broadway.Message.t,\n        kind: kind,\n        reason: reason,\n        stacktrace: stacktrace,\n        telemetry_span_context: reference\n      }\n      ```\n\n  * `[:broadway, :batch_processor, :start]` - Dispatched by a Broadway batch processor\n    before your `c:handle_batch/4` callback is invoked\n\n    * Measurement: `%{system_time: integer}`\n    * Metadata:\n\n      ```\n      %{\n        topology_name: atom,\n        name: atom,\n        index: non_neg_integer,\n        messages: [Broadway.Message.t],\n        batch_info: Broadway.BatchInfo.t,\n        telemetry_span_context: reference,\n        producer: {atom, list}\n      }\n      ```\n\n  * `[:broadway, :batch_processor, :stop]` - Dispatched by a Broadway batch\n    processor after your `c:handle_batch/4` callback has returned\n\n    * Measurement: `%{duration: native_time}`\n\n    * Metadata:\n\n      ```\n      %{\n        topology_name: atom,\n        name: atom,\n        index: non_neg_integer,\n        successful_messages: [Broadway.Message.t],\n        failed_messages: [Broadway.Message.t],\n        batch_info: Broadway.BatchInfo.t,\n        telemetry_span_context: reference,\n        producer: {atom, list}\n      }\n      ```\n\n  * `[:broadway, :batcher, :start]` - Dispatched by a Broadway batcher before\n    handling events\n\n    * Measurement: `%{system_time: integer}`\n    * Metadata:\n\n      ```\n      %{\n        topology_name: atom,\n        name: atom,\n        batcher_key: atom,\n        messages: [Broadway.Message.t],\n        telemetry_span_context: reference\n      }\n      ```\n\n  * `[:broadway, :batcher, :stop]` - Dispatched by a Broadway batcher after\n    handling events\n\n    * Measurement: `%{duration: native_time}`\n    * Metadata:\n\n    ```\n      %{\n        topology_name: atom,\n        name: atom,\n        batcher_key: atom,\n        telemetry_span_context: reference\n      }\n    ```\n\nMost of the events follow the `:telemetry.span/3` convention for measurements.\nThis means that \"start\" events have a `:system_time` representing the start of\nthat event using `System.system_time/0`. The \"stop\" or \"exception\" events\nhave the `duration` value, which is the difference in monotonic time between\nthe start and stop events.","ref":"Broadway.html#module-telemetry"},{"type":"function","title":"Broadway.all_running/0","doc":"Returns all running Broadway names.\n\nIt's important to notice that no order is guaranteed.","ref":"Broadway.html#all_running/0"},{"type":"callback","title":"Broadway.format_discarded/2","doc":"Invoked when items are discarded from the buffer.\n\nIf this callback returns `true`, the default log message is emitted.\nSee `c:GenStage.format_discarded/2`.\n\nAllows controlling or customization of the log message emitted.","ref":"Broadway.html#c:format_discarded/2"},{"type":"function","title":"Broadway.get_rate_limiting/1","doc":"Gets the current values used for the producer rate limiting of the given pipeline.\n\nReturns `{:ok, info}` if rate limiting is enabled for the given pipeline or\n`{:error, reason}` if the given pipeline doesn't have rate limiting enabled.\n\nThe returned info is a map with the following keys:\n\n  * `:interval`\n  * `:allowed_messages`\n\nSee the `:rate_limiting` options in the module documentation for more information.","ref":"Broadway.html#get_rate_limiting/1"},{"type":"function","title":"Examples - Broadway.get_rate_limiting/1","doc":"Broadway.get_rate_limiting(broadway)\n    #=> {:ok, %{allowed_messages: 2000, interval: 1000}}","ref":"Broadway.html#get_rate_limiting/1-examples"},{"type":"callback","title":"Broadway.handle_batch/4","doc":"Invoked to handle generated batches.\n\nIt expects:\n\n  * `batcher` is the key that defined the batcher. This value can be\n    set in the `handle_message/3` callback using `Broadway.Message.put_batcher/2`.\n  * `messages` is the list of `Broadway.Message` structs in the incoming batch.\n  * `batch_info` is a `Broadway.BatchInfo` struct containing extra information\n    about the incoming batch.\n  * `context` is the user defined data structure passed to `start_link/2`.\n\nIt must return an updated list of messages. All messages received must be returned,\notherwise an error will be logged. All messages after this step will be acknowledged\naccording to their status.\n\nIn case of errors in this callback, the error will be logged and the whole\nbatch will be failed. This callback also traps exits, so failures due to broken\nlinks between processes do not automatically cascade.\n\nFor more information on batching, see the \"Batching\" section in the `Broadway`\ndocumentation.","ref":"Broadway.html#c:handle_batch/4"},{"type":"callback","title":"Broadway.handle_failed/2","doc":"Invoked for failed messages (if defined).\n\nIt expects:\n\n  * `messages` is the list of messages that failed. If a message is failed in\n    `c:handle_message/3`, this will be a list with a single message in it. If\n    some messages are failed in `c:handle_batch/4`, this will be the list of\n    failed messages.\n\n  * `context` is the user-defined data structure passed to `start_link/2`.\n\nThis callback must return the same messages given to it, possibly updated.\nFor example, you could update the message data or use `Broadway.Message.configure_ack/2`\nin a centralized place to configure how to ack the message based on the failure\nreason.\n\nThis callback is optional. If present, it's called **before** the messages\nare acknowledged according to the producer. This gives you a chance to do something\nwith the message before it's acknowledged, such as storing it in an external\npersistence layer or similar.\n\nThis callback is also invoked if `c:handle_message/3` or `c:handle_batch/4`\ncrash or raise an error. If this callback crashes or raises an error,\nthe messages are failed internally by Broadway to avoid crashing the process.","ref":"Broadway.html#c:handle_failed/2"},{"type":"callback","title":"Broadway.handle_message/3","doc":"Invoked to handle/process individual messages sent from a producer.\n\nIt receives:\n\n  * `processor`  is the key that defined the processor.\n  * `message` is the `Broadway.Message` struct to be processed.\n  * `context` is the user defined data structure passed to `start_link/2`.\n\nAnd it must return the (potentially) updated `Broadway.Message` struct.\n\nThis is the place to do any kind of processing with the incoming message,\ne.g., transform the data into another data structure, call specific business\nlogic to do calculations. Basically, any CPU bounded task that runs against\na single message should be processed here.\n\nIn order to update the data after processing, use the\n`Broadway.Message.update_data/2` function. This way the new message can be\nproperly forwarded and handled by the batcher:\n\n    @impl true\n    def handle_message(_, message, _) do\n      message\n      |> update_data(&do_calculation_and_returns_the_new_data/1)\n    end\n\nIn case more than one batcher have been defined in the configuration,\nyou need to specify which of them the resulting message will be forwarded\nto. You can do this by calling `put_batcher/2` and returning the new\nupdated message:\n\n    @impl true\n    def handle_message(_, message, _) do\n      # Do whatever you need with the data\n      ...\n\n      message\n      |> put_batcher(:s3)\n    end\n\nAny message that has not been explicitly failed will be forwarded to the next\nstep in the pipeline. If there are no extra steps, it will be automatically\nacknowledged.\n\nIn case of errors in this callback, the error will be logged and that particular\nmessage will be immediately acknowledged as failed, not proceeding to the next\nsteps of the pipeline. This callback also traps exits, so failures due to broken\nlinks between processes do not automatically cascade.","ref":"Broadway.html#c:handle_message/3"},{"type":"type","title":"Broadway.name/0","doc":"","ref":"Broadway.html#t:name/0"},{"type":"type","title":"Broadway.on_start/0","doc":"Returned by `start_link/2`.","ref":"Broadway.html#t:on_start/0"},{"type":"callback","title":"Broadway.prepare_messages/2","doc":"Invoked for preparing messages before handling (if defined).\n\nIt expects:\n\n  * `messages` is a list of `Broadway.Message` structs to be processed.\n  * `context` is the user defined data structure passed to `start_link/2`.\n\nThis is the place to prepare and preload any information that will be used\nby `c:handle_message/3`. For example, if you need to query the database,\ninstead of doing it once per message, you can do it on this callback as\na best-effort optimization.\n\nThe length of the list of messages received by this callback is often based\non the `min_demand`/`max_demand` configuration in the processor but ultimately\nit depends on the producer and on the frequency data arrives. A pipeline that\nreceives messages rarely will most likely emit lists of length below the\n`min_demand` value. Producers which are push-based, rather than pull-based,\nsuch as `BroadwayRabbitMQ.Producer`, are more likely to send messages as they\narrive (which may skip batching altogether and always be single element lists).\nIn other words, this callback is simply a convenience for preparing messages,\nit does not guarantee the messages will be accumulated to a certain length.\nFor effective batch processing, see `c:handle_batch/4`.\n\nThis callback must always return all messages it receives, as\n`c:handle_message/3` is still called individually for each message afterwards.\n\n> #### Failed Messages {: .warning}\n>\n> Even if `c:prepare_messages/2` **fails** some messages (`Broadway.Message.failed/2`),\n> the failed messages are still passed down to `c:handle_message/3`.\n> If your pipeline wants to avoid processing messages failed in `c:prepare_messages/2`,\n> it will have to pattern match on `%Broadway.Message{status: {:failed, reason}}`\n> in its `c:handle_message/3` callback and act accordingly.","ref":"Broadway.html#c:prepare_messages/2"},{"type":"callback","title":"Broadway.process_name/2","doc":"Invoked to get the process name of this Broadway pipeline.\n\n`broadway_name` is the name given to `start_link/2` in the `:name` option. `base_name`\nis a string used by Broadway to identify different components of the pipeline\nwhose name needs to be registered (such as \"batcher\" or \"processor\").\n\nThe return value of this callback must be a process name that is valid for registration.\nSee the name registration rules in the documentation for `GenServer`.\n\nThis callback is optional. If not defined, the `broadway_name` given to `start_link/2`\n**must be an atom**: the default implementation of this callback will fail otherwise.","ref":"Broadway.html#c:process_name/2"},{"type":"callback","title":"Examples - Broadway.process_name/2","doc":"@impl Broadway\n    def process_name({:via, module, term}, base_name) do\n      {:via, module, {term, base_name}}\n    end","ref":"Broadway.html#c:process_name/2-examples"},{"type":"function","title":"Broadway.producer_names/1","doc":"Returns the names of producers.","ref":"Broadway.html#producer_names/1"},{"type":"function","title":"Examples - Broadway.producer_names/1","doc":"iex> Broadway.producer_names(MyBroadway)\n    [MyBroadway.Producer_0, MyBroadway.Producer_1, ..., MyBroadway.Producer_7]","ref":"Broadway.html#producer_names/1-examples"},{"type":"function","title":"Broadway.push_messages/2","doc":"Sends a list of `Broadway.Message`s to the Broadway pipeline.\n\nThe producer is randomly chosen among all sets of producers/stages.\nThis is used to send out of band data to a Broadway pipeline.","ref":"Broadway.html#push_messages/2"},{"type":"function","title":"Broadway.start_link/2","doc":"Starts a `Broadway` process linked to the current process.\n\n  * `module` is the module implementing the `Broadway` behaviour.","ref":"Broadway.html#start_link/2"},{"type":"function","title":"Options - Broadway.start_link/2","doc":"In order to set up how the pipeline created by Broadway should work,\nyou need to specify the blueprint of the pipeline. You can\ndo this by passing a set of options to `start_link/2`.\nEach component of the pipeline has its own set of options.\n\nThe Broadway options are:\n\n* `:name` - Required. Used for name registration. When an atom, all processes/stages\n  created will be named using this value as prefix.\n\n* `:shutdown` (`t:pos_integer/0`) - Optional. The time in milliseconds given for Broadway to\n  gracefully shutdown without discarding events. The default value is `30000`.\n\n* `:max_restarts` (`t:non_neg_integer/0`) - The default value is `3`.\n\n* `:max_seconds` (`t:pos_integer/0`) - The default value is `5`.\n\n* `:resubscribe_interval` (`t:non_neg_integer/0`) - The interval in milliseconds that\n  processors wait until they resubscribe to a failed producers. The default value is `100`.\n\n* `:context` (`t:term/0`) - A user defined data structure that will be passed to handle_message/3 and handle_batch/4. The default value is `:context_not_set`.\n\n* `:producer` (non-empty `t:keyword/0`) - Required. A keyword list of options. See [\"Producers options\"](#start_link/2-producers-options)\n  section below. Only a single producer is allowed.\n\n* `:processors` (non-empty `t:keyword/0`) - Required. A keyword list of named processors where the key is an atom as identifier and\n  the value is another keyword list of options.\n  See [\"Processors options\"](#start_link/2-processors-options)\n  section below. Currently only a single processor is allowed.\n\n* `:batchers` (`t:keyword/0`) - A keyword list of named batchers\n  where the key is an atom as identifier and the value is another\n  keyword list of options. See [\"Batchers options\"](#start_link/2-batchers-options)\n  section below. The default value is `[]`.\n\n* `:partition_by` (function of arity 1) - A function that controls how data is\n  partitioned across all processors and batchers. It receives a\n  `Broadway.Message` and it must return a non-negative integer,\n  starting with zero, that will be mapped to one of the existing\n  processors. See [\"Ordering and Partitioning\"](#module-ordering-and-partitioning)\n  in the module docs for more information and known pitfalls.\n\n* `:spawn_opt` (`t:keyword/0`) - Low-level options given when starting a\n  process. Applies to producers, processors, and batchers.\n  See `erlang:spawn_opt/2` for more information.\n\n* `:hibernate_after` (`t:pos_integer/0`) - If a process does not receive any message within this interval, it will hibernate,\n  compacting memory. Applies to producers, processors, and batchers.\n  Defaults to `15_000` (millisecond). The default value is `15000`.","ref":"Broadway.html#start_link/2-options"},{"type":"function","title":"Producers options - Broadway.start_link/2","doc":"The producer options allow users to set up the producer.\n\nThe available options are:\n\n* `:module` - Required. A tuple representing a GenStage producer.\n  The tuple format should be `{mod, arg}`, where `mod` is the module\n  that implements the GenStage behaviour and `arg` the argument that will\n  be passed to the `init/1` callback of the producer. See `Broadway.Producer`\n  for more information.\n\n* `:concurrency` (`t:pos_integer/0`) - The number of concurrent producers that\n  will be started by Broadway. Use this option to control the concurrency\n  level of each set of producers. The default value is `1`.\n\n* `:transformer` - A tuple representing a transformer that translates a produced GenStage event into a\n  `%Broadway.Message{}`. The tuple format should be `{mod, fun, opts}` and the function\n  should have the following spec `(event :: term, opts :: term) :: Broadway.Message.t`\n  This function must be used sparingly and exclusively to convert regular\n  messages into `Broadway.Message`. That's because a failure in the\n  `:transformer` callback will cause the whole producer to terminate,\n  possibly leaving unacknowledged messages along the way. The default value is `nil`.\n\n* `:spawn_opt` (`t:keyword/0`) - Overrides the top-level `:spawn_opt`.\n\n* `:hibernate_after` (`t:pos_integer/0`) - Overrides the top-level `:hibernate_after`.\n\n* `:rate_limiting` (non-empty `t:keyword/0`) - A list of options to enable and configure rate limiting for producing.\n  If this option is present, rate limiting is enabled, otherwise it isn't.\n  Rate limiting refers to the rate at which producers will forward\n  messages to the rest of the pipeline. The rate limiting is applied to\n  and shared by all producers within the time limit.\n  The following options are supported:\n\n  * `:allowed_messages` (`t:pos_integer/0`) - Required. An integer that describes how many messages are allowed in the specified interval.\n\n  * `:interval` (`t:pos_integer/0`) - Required. An integer that describes the interval (in milliseconds)\n    during which the number of allowed messages is allowed.\n    If the producer produces more than `allowed_messages`\n    in `interval`, only `allowed_messages` will be published until\n    the end of `interval`, and then more messages will be published.","ref":"Broadway.html#start_link/2-producers-options"},{"type":"function","title":"Processors options - Broadway.start_link/2","doc":"> #### You don't need multiple processors {: .info}\n>\n> A common misconception is that, if your data requires multiple\n> transformations, each with a different concern, then you must\n> have several processors.\n>\n> However, that's not quite true. Separation of concerns is modeled\n> by defining several modules and functions, not processors. Processors\n> are ultimately about moving data around and you should only do it\n> when necessary. Using processors for code organization purposes would\n> lead to inefficient pipelines.\n\n* `:concurrency` (`t:pos_integer/0`) - The number of concurrent process that will\n  be started by Broadway. Use this option to control the concurrency level\n  of the processors. The default value is `System.schedulers_online() * 2`.\n\n* `:min_demand` (`t:non_neg_integer/0`) - Set the minimum demand of all processors stages.\n\n* `:max_demand` (`t:non_neg_integer/0`) - Set the maximum demand of all processors stages. The default value is `10`.\n\n* `:partition_by` (function of arity 1) - Overrides the top-level `:partition_by`.\n\n* `:spawn_opt` (`t:keyword/0`) - Overrides the top-level `:spawn_opt`.\n\n* `:hibernate_after` (`t:pos_integer/0`) - Overrides the top-level `:hibernate_after`.","ref":"Broadway.html#start_link/2-processors-options"},{"type":"function","title":"Batchers options - Broadway.start_link/2","doc":"* `:concurrency` (`t:pos_integer/0`) - The number of concurrent batch processors\n  that will be started by Broadway. Use this option to control the\n  concurrency level. Note that this only sets the numbers of batch\n  processors for each batcher group, not the number of batchers.\n  The number of batchers will always be one for each batcher key\n  defined. The default value is `1`.\n\n* `:batch_size` - The size of the generated batches. Default value is `100`. It is typically an\n  integer but it can also be tuple of `{init_acc, fun}`\n  where `fun` receives two arguments: a `Broadway.Message` and\n  an `acc`. The function must return either `{:emit, acc}` to indicate\n  all batched messages must be emitted or `{:cont, acc}` to continue\n  batching. `init_acc` is the initial accumulator used on the first call. You can\n  consider that setting the accumulator to an integer is the equivalent to custom\n  batching function of:\n\n      {batch_size,\n       fn\n         _message, 1 -> {:emit, batch_size}\n         _message, count -> {:cont, count - 1}\n       end}\n\n  The default value is `100`.\n\n* `:max_demand` (`t:pos_integer/0`) - Sets the maximum demand of batcher stages.\n  By default it is set to `:batch_size`, if `:batch_size` is an integer.\n  Must be set if the `:batch_size` is a function.\n\n* `:batch_timeout` (`t:pos_integer/0`) - The time, in milliseconds, that the batcher waits before flushing\n  the list of messages. When this timeout is reached, a new batch\n  is generated and sent downstream, no matter if the `:batch_size`\n  has been reached or not. The default value is `1000`.\n\n* `:partition_by` (function of arity 1) - Optional. Overrides the top-level `:partition_by`.\n\n* `:spawn_opt` (`t:keyword/0`) - Overrides the top-level `:spawn_opt`.\n\n* `:hibernate_after` (`t:pos_integer/0`) - Overrides the top-level `:hibernate_after`.","ref":"Broadway.html#start_link/2-batchers-options"},{"type":"function","title":"Broadway.stop/3","doc":"Synchronously stops the Broadway pipeline with the given `reason`.\n\nThis function returns `:ok` if the pipeline terminates with the\ngiven reason; if it terminates with another reason, the call exits.\n\nThis function keeps OTP semantics regarding error reporting.\nIf the reason is any other than `:normal`, `:shutdown` or\n`{:shutdown, _}`, an error report is logged.","ref":"Broadway.html#stop/3"},{"type":"function","title":"Broadway.test_batch/3","doc":"Sends a list of data as a batch of messages to the Broadway pipeline.\n\nThis is a convenience used for testing. Each message is automatically\nwrapped in a `Broadway.Message` with `Broadway.CallerAcknowledger`\nconfigured to send a message back to the caller once all batches\nhave been fully processed.\n\nIf there are more messages in the batch than the pipeline `batch_size`\nor if the messages in the batch take more time to process than\n`batch_timeout` then the caller will receive multiple messages.\n\nIt returns a reference that can be used to identify the ack\nmessages.\n\nSee [\"Testing\"](#module-testing) section in module documentation\nfor more information.","ref":"Broadway.html#test_batch/3"},{"type":"function","title":"Options - Broadway.test_batch/3","doc":"* `:metadata` (`t:term/0`) - an enumerable of key-value pairs of *additional* fields to add to the\n  message. This can be used, for example, when testing `BroadwayRabbitMQ.Producer`. The default value is `[]`.\n\n* `:acknowledger` (function of arity 2) - a function that generates `ack` fields of the sent `Broadway.Message.t()`.\n  This function receives the acknowledger `data` and the `from` field and\n  it must return the acknowledger tuple. The typespec of this function is:\n\n  `data :: term(), from :: {pid(), term()} -> {module(), ack_ref :: term(), ack_data :: term()}`\n\n* `:batch_mode` - when set to `:flush`, the batch the message is in is immediately delivered. When set\n  to `:bulk`, batch is delivered when its size or timeout is reached. The default value is `:bulk`.","ref":"Broadway.html#test_batch/3-options"},{"type":"function","title":"Examples - Broadway.test_batch/3","doc":"For example, in your tests, you may do:\n\n    ref = Broadway.test_batch(broadway, [1, 2, 3])\n    assert_receive {:ack, ^ref, successful, failed}, 1000\n    assert length(successful) == 3\n    assert length(failed) == 0\n\nNote that messages sent using this function will ignore demand and :transform\noption specified in :producer option in `Broadway.start_link/2`.","ref":"Broadway.html#test_batch/3-examples"},{"type":"function","title":"Broadway.test_message/3","doc":"Sends a test message through the Broadway pipeline.\n\nThis is a convenience used for testing. The given data\nis automatically wrapped in a `Broadway.Message` with\n`Broadway.CallerAcknowledger` configured to send a message\nback to the caller once the message has been fully processed.\n\nThe message is set to be flushed immediately, without waiting\nfor the Broadway pipeline `batch_size` to be filled or the\n`batch_timeout` to be triggered.\n\nIt returns a reference that can be used to identify the ack\nmessages.\n\nSee [\"Testing\"](#module-testing) section in module documentation\nfor more information.","ref":"Broadway.html#test_message/3"},{"type":"function","title":"Options - Broadway.test_message/3","doc":"* `:metadata` (`t:term/0`) - an enumerable of key-value pairs of *additional* fields to add to the\n  message. This can be used, for example, when testing `BroadwayRabbitMQ.Producer`. The default value is `[]`.\n\n* `:acknowledger` (function of arity 2) - a function that generates `ack` fields of the sent `Broadway.Message.t()`.\n  This function receives the acknowledger `data` and the `from` field and\n  it must return the acknowledger tuple. The typespec of this function is:\n\n  `data :: term(), from :: {pid(), term()} -> {module(), ack_ref :: term(), ack_data :: term()}`","ref":"Broadway.html#test_message/3-options"},{"type":"function","title":"Examples - Broadway.test_message/3","doc":"For example, in your tests, you may do:\n\n    ref = Broadway.test_message(broadway, 1)\n    assert_receive {:ack, ^ref, [successful], []}\n\nor if you want to override which acknowledger shall be called, you may do:\n\n    acknowledger = fn data, ack_ref -> {MyAck, ack_ref, :ok} end\n    Broadway.test_message(broadway, 1, acknowledger: acknowledger)\n\nNote that messages sent using this function will ignore demand and :transform\noption specified in :producer option in `Broadway.start_link/2`.","ref":"Broadway.html#test_message/3-examples"},{"type":"function","title":"Broadway.topology/1","doc":"Returns the topology details for a pipeline.\n\nThe stages that have the `:concurrency` field in their info indicate a list of\nprocesses running with that name prefix. Each process has `:name` as a\nprefix plus `_` and the index of `0..(concurrency - 1)` as an atom. For example, a\nproducer named `MyBroadway.Broadway.Producer` with concurrency of `1`\nwill only have a single process named `MyBroadway.Broadway.Producer_0` in its\ntopology.\n\n> #### Single producer and processor {: .info}\n>\n> Broadway does not accept multiple producers neither\n> multiple processors, but we chose to keep in a list\n> for consistency and to ensure we're future proof.","ref":"Broadway.html#topology/1"},{"type":"function","title":"Examples - Broadway.topology/1","doc":"iex> Broadway.topology(MyBroadway)\n    [\n      producers: [\n        %{name: MyBroadway.Broadway.Producer, concurrency: 1}\n      ],\n      processors: [\n        %{name: MyBroadway.Broadway.Processor_default, concurrency: 10, processor_key: :default}\n      ],\n      batchers: [\n        %{\n          batcher_name: MyBroadway.Broadway.Batcher_default,\n          name: MyBroadway.Broadway.BatchProcessor_default,\n          batcher_key: :default,\n          concurrency: 5\n        },\n        %{\n          batcher_name: MyBroadway.Broadway.Batcher_s3,\n          name: MyBroadway.Broadway.BatchProcessor_s3,\n          batcher_key: :s3,\n          concurrency: 3\n        }\n      ]\n    ]\n\nIn the example above, for instance, the processor process names would be\n`MyBroadway.Broadway.Processor_default_0`, `MyBroadway.Broadway.Processor_default_1`,\nand so on.","ref":"Broadway.html#topology/1-examples"},{"type":"function","title":"Broadway.update_rate_limiting/2","doc":"Updates the producer rate limiting of the given pipeline at runtime.\n\nSupports the following options (see the `:rate_limiting` options in the module\ndocumentation for more information):\n\n  * `:allowed_messages`\n  * `:interval`\n  * `:reset`\n\nReturns an `{:error, reason}` tuple if the given `broadway` pipeline doesn't\nhave rate limiting enabled.\n\nThe option `:reset` defaults to `false`. This means the rate limit will reset\nto the new rate limit at the end of the current interval. When `:reset` is `true`,\nthe new rate limit takes effect immediately.","ref":"Broadway.html#update_rate_limiting/2"},{"type":"function","title":"Examples - Broadway.update_rate_limiting/2","doc":"Broadway.update_rate_limiting(broadway, allowed_messages: 100)","ref":"Broadway.html#update_rate_limiting/2-examples"},{"type":"module","title":"Broadway.BatchInfo","doc":"A struct used to hold information about a generated batch.\n\nAn instance of this struct containing the related info will\nbe passed to the `c:Broadway.handle_batch/4` callback of the\nmodule implementing the `Broadway` behaviour.\n\nSee the documentation for [`%Broadway.BatchInfo{}`](`__struct__/0`)\nfor information on the fields.","ref":"Broadway.BatchInfo.html"},{"type":"function","title":"Broadway.BatchInfo.__struct__/0","doc":"The batch info struct.\n\nThe fields are:\n\n  * `:batcher` - is the key that defined the batcher. This value can\n    be set in the `c:Broadway.handle_message/3` callback using\n    `Broadway.Message.put_batcher/2`.\n\n  * `:batch_key` - identifies the batch key for this batch.\n    See `Broadway.Message.put_batch_key/2`.\n\n  * `:partition` - the partition, if present.\n\n  * `:size` - the number of messages in the batch.\n\n  * `:trigger` - the trigger that generated the batch, like `:timeout`\n    or `:flush`.","ref":"Broadway.BatchInfo.html#__struct__/0"},{"type":"type","title":"Broadway.BatchInfo.t/0","doc":"The type for a batch info struct.","ref":"Broadway.BatchInfo.html#t:t/0"},{"type":"module","title":"Broadway.Message","doc":"This struct holds all information about a message.\n\nA message is first created by the producers. It is then\nsent downstream and gets updated multiple times, either\nby a module implementing the `Broadway` behaviour\nthrough the `c:Broadway.handle_message/3` callback\nor internally by one of the built-in stages of Broadway.\n\nInstead of modifying the struct directly, you should use the functions\nprovided by this module to manipulate messages. However, if you are implementing\na `Broadway.Producer` of your own, see `t:t/0` to see what fields you should set.","ref":"Broadway.Message.html"},{"type":"function","title":"Broadway.Message.ack_immediately/1","doc":"Immediately acknowledges the given message or list of messages.\n\nThis function can be used to acknowledge a message (or list of messages)\nimmediately without waiting for the rest of the pipeline.\n\nAcknowledging a message sets that message's acknowledger to a no-op\nacknowledger so that it's safe to ack at the end of the pipeline.\n\nReturns the updated acked message if a message is passed in,\nor the updated list of acked messages if a list of messages is passed in.","ref":"Broadway.Message.html#ack_immediately/1"},{"type":"type","title":"Broadway.Message.acknowledger/0","doc":"The acknowledger of the message.\n\nThis tuple contains:\n\n  * A module implementing the `Broadway.Acknowledger` behaviour.\n\n  * An ack reference that is passed to the `c:Broadway.Acknowledger.ack/3`\n    callback. See `c:Broadway.Acknowledger.ack/3` for more information.\n\n  * An arbitrary term that is passed to the optional\n    `c:Broadway.Acknowledger.configure/3` callback.","ref":"Broadway.Message.html#t:acknowledger/0"},{"type":"function","title":"Broadway.Message.configure_ack/2","doc":"Configures the acknowledger of this message.\n\nThis function calls the `c:Broadway.Acknowledger.configure/3` callback to\nchange the configuration of the acknowledger for the given `message`.\n\nThis function can only be called if the acknowledger implements the `configure/3`\ncallback. If it doesn't, an error is raised.","ref":"Broadway.Message.html#configure_ack/2"},{"type":"function","title":"Broadway.Message.failed/2","doc":"Mark a message as failed.\n\nFailed messages are sent directly to the related acknowledger at the end\nof this step and therefore they're not forwarded to the next step in the\npipeline.\n\nFailing a message does not emit any log but it does trigger the\n`c:Broadway.handle_failed/2` callback.","ref":"Broadway.Message.html#failed/2"},{"type":"function","title":"Broadway.Message.put_batch_key/2","doc":"Defines the message batch key.\n\nThe batch key identifies the batch the message belongs to, within\na given batcher. Each batcher then groups batches with the same\n`batch_key`, with size of at most `batch_size` within period\n`batch_timeout`. Both `batch_size` and `batch_timeout` are managed\nper batch key, so a batcher is capable of grouping multiple batch\nkeys at the same time, regardless of the concurrency level.\n\nIf a given batcher has multiple batch processors (concurrency > 1),\nall messages with the same batch key are routed to the same processor.\nSo different batch keys may run concurrently but the same batch key\nis always run serially and in the same batcher processor.","ref":"Broadway.Message.html#put_batch_key/2"},{"type":"function","title":"Broadway.Message.put_batch_mode/2","doc":"Sets the batching mode for the message.\n\nWhen the mode is `:bulk`, the batch that the message is in is delivered after\nthe batch size or batch timeout is reached.\n\nWhen the mode is `:flush`, the batch that the message is in is delivered\nimmediately after processing. Note it doesn't mean the batch contains only a single element\nbut rather that all messages received from the processor are delivered without waiting.\n\nThe default mode for messages is `:bulk`.","ref":"Broadway.Message.html#put_batch_mode/2"},{"type":"function","title":"Broadway.Message.put_batcher/2","doc":"Defines the target batcher which the message should be forwarded to.","ref":"Broadway.Message.html#put_batcher/2"},{"type":"function","title":"Broadway.Message.put_data/2","doc":"Stores the given data in the message.\n\nThis function is usually used inside the `c:Broadway.handle_message/3` implementation\nto replace data with new processed data.","ref":"Broadway.Message.html#put_data/2"},{"type":"type","title":"Broadway.Message.t/0","doc":"The Broadway message struct.\n\nMost of these fields are manipulated by Broadway itself. You can\n*read* the `:metadata` field, and you can use the functions in this\nmodule to update most of the other fields. If you are implementing\nyour own producer, see the `Broadway.Producer` documentation\nfor more information on how to create and manipulate message structs.","ref":"Broadway.Message.html#t:t/0"},{"type":"function","title":"Broadway.Message.update_data/2","doc":"Updates the data in the message.\n\nThis function is usually used inside the `c:Broadway.handle_message/3` implementation\nto update data with new processed data.","ref":"Broadway.Message.html#update_data/2"},{"type":"behaviour","title":"Broadway.Acknowledger","doc":"A behaviour used to acknowledge that the received messages\nwere successfully processed or failed.\n\nWhen implementing a new connector for Broadway, you should\nimplement this behaviour and consider how the technology\nyou're working with handles message acknowledgement.\n\nThe `c:ack/3` callback must be implemented in order to notify\nthe origin of the data that a message can be safely removed\nafter been successfully processed and published. In case of\nfailed messages or messages without acknowledgement, depending\non the technology chosen, the messages can be either moved back\nin the queue or, alternatively, moved to a *dead-letter queue*.","ref":"Broadway.Acknowledger.html"},{"type":"callback","title":"Broadway.Acknowledger.ack/3","doc":"Invoked to acknowledge successful and failed messages.\n\n  * `ack_ref` is a term that uniquely identifies how messages\n    should be grouped and sent for acknowledgement. Imagine\n    you have a scenario where messages are coming from\n    different producers. Broadway will use this information\n    to correctly identify the acknowledger and pass it among\n    with the messages so you can properly communicate with\n    the source of the data for acknowledgement. `ack_ref` is\n    part of `t:Broadway.Message.acknowledger/0`.\n\n  * `successful` is the list of messages that were\n    successfully processed and published.\n\n  * `failed` is the list of messages that, for some reason,\n    could not be processed or published.","ref":"Broadway.Acknowledger.html#c:ack/3"},{"type":"callback","title":"Broadway.Acknowledger.configure/3","doc":"Configures the acknowledger with new `options`.\n\nEvery acknowledger can decide how to incorporate the given `options` into its\n`ack_data`. The `ack_data` is the current acknowledger's data. The return value\nof this function is `{:ok, new_ack_data}` where `new_ack_data` is the updated\ndata for the acknowledger.\n\nNote that `options` are different for every acknowledger, as the acknowledger\nis what specifies what are the supported options. Check the documentation for the\nacknowledger you're using to see the supported options.\n\n`ack_ref` and `ack_data` are part of `t:Broadway.Message.acknowledger/0`.","ref":"Broadway.Acknowledger.html#c:configure/3"},{"type":"module","title":"Broadway.CallerAcknowledger","doc":"A simple acknowledger that sends a message back to a caller.\n\nIf you want to use this acknowledger in messages produced by your\n`Broadway.Producer`, you can get its configuration by calling\nthe `init/0` function. For example, you can use it in\n`Broadway.test_message/3`:\n\n    some_ref = make_ref()\n\n    Broadway.test_message(\n      MyPipeline,\n      \"some data\",\n      acknowledger: Broadway.CallerAcknowledger.init({self(), some_ref}, :ignored)\n    )\n\nThe first parameter is a tuple with the PID to receive the messages\nand a unique identifier (usually a reference). Such unique identifier\nis then included in the messages sent to the PID. The second parameter,\nwhich is per message, is ignored.\n\nIt sends a message in the format:\n\n    {:ack, ref, successful_messages, failed_messages}\n\nIf `Broadway.Message.configure_ack/2` is called on a message that\nuses this acknowledger, then the following message is sent:\n\n    {:configure, ref, options}","ref":"Broadway.CallerAcknowledger.html"},{"type":"function","title":"Broadway.CallerAcknowledger.init/2","doc":"Returns the acknowledger metadata.\n\nSee the module documentation.","ref":"Broadway.CallerAcknowledger.html#init/2"},{"type":"module","title":"Broadway.NoopAcknowledger","doc":"An acknowledger that does nothing.\n\nIf you want to use this acknowledger in messages produced by your\n`Broadway.Producer`, you can get its configuration by calling\nthe `init/0` function. For example, you can use it in\n`Broadway.test_message/3`:\n\n    Broadway.test_message(MyPipeline, \"some data\", acknowledger: Broadway.NoopAcknowledger.init())\n\nBroadway sets this acknowledger automatically on messages that have been acked\nvia `Broadway.Message.ack_immediately/1`.","ref":"Broadway.NoopAcknowledger.html"},{"type":"function","title":"Broadway.NoopAcknowledger.init/0","doc":"Returns the acknowledger metadata.","ref":"Broadway.NoopAcknowledger.html#init/0"},{"type":"module","title":"Broadway.DummyProducer","doc":"A producer that does nothing, used mostly for testing.\n\nSee \"Testing\" section in `Broadway` module documentation for more information.","ref":"Broadway.DummyProducer.html"},{"type":"behaviour","title":"Broadway.Producer","doc":"A Broadway producer is a `GenStage` producer that emits\n`Broadway.Message` structs as events.\n\nThe `Broadway.Producer` is declared in a Broadway topology\nvia the `:module` option (see `Broadway.start_link/2`):\n\n    producer: [\n      module: {MyProducer, options}\n    ]\n\nOnce declared, `MyProducer` is expected to implement and\nbehave as a `GenStage` producer. When Broadway starts,\nthe `c:GenStage.init/1` callback will be invoked directly with the\ngiven `options`.","ref":"Broadway.Producer.html"},{"type":"behaviour","title":"Injected Broadway configuration - Broadway.Producer","doc":"If `options` is a keyword list, Broadway injects a `:broadway` option\ninto such keyword list. This option contains the configuration for the\ncomplete Broadway topology (see `Broadway.start_link/2`. For example,\nyou can use `options[:broadway][:name]` to uniquely identify the topology.\n\nThe `:broadway` configuration also has an `:index` key. This\nis the index of the producer in its supervision tree (starting\nfrom `0`). This allows a features such having even producers\nconnect to some server while odd producers connect to another.\n\nIf `options` is any other term, it is passed as is to the `c:GenStage.init/1`\ncallback. All other functions behave precisely as in `GenStage`\nwith the requirements that all emitted events must be `Broadway.Message`\nstructs.","ref":"Broadway.Producer.html#module-injected-broadway-configuration"},{"type":"behaviour","title":"Optional callbacks - Broadway.Producer","doc":"A `Broadway.Producer` can implement two optional Broadway callbacks,\n`c:prepare_for_start/2` and `c:prepare_for_draining/1`, which are useful\nfor booting up and shutting down Broadway topologies respectively.","ref":"Broadway.Producer.html#module-optional-callbacks"},{"type":"behaviour","title":"Producing Broadway messages - Broadway.Producer","doc":"You should generally modify `Broadway.Message` structs by using the functions\nin the `Broadway.Message` module. However, if you are implementing your\nown producer, you **can manipulate** some of the struct's fields directly.\n\nThese fields are:\n\n  * `:data` (required) - the data of the message. Even though the function\n    `Broadway.Message.put_data/2` exists, when creating a `%Broadway.Message{}`\n    struct from scratch you will have to pass in the `:data` field directly.\n\n  * `:acknowledger` (required) - the acknowledger of the message, of type\n    `t:Broadway.Message.acknowledger/0`.\n\n  * `:metadata` (optional) - metadata about the message that your producer\n    can attach to the message. This is useful when you want to add some metadata\n    to messages, and document it for users to use in their pipelines.\n\nFor example, a producer could create a message by doing something like this:\n\n    %Broadway.Message{\n      data: \"some data here\",\n      acknowledger: Broadway.NoopAcknowledger.init()\n    }","ref":"Broadway.Producer.html#module-producing-broadway-messages"},{"type":"callback","title":"Broadway.Producer.prepare_for_draining/1","doc":"Invoked by the terminator right before Broadway starts draining in-flight\nmessages during shutdown.\n\nThis callback should be implemented by producers that need to do additional\nwork before shutting down. That includes active producers like RabbitMQ that\nmust ask the data provider to stop sending messages. It will be invoked for\neach producer stage.\n\n`state` is the current state of the producer.","ref":"Broadway.Producer.html#c:prepare_for_draining/1"},{"type":"callback","title":"Broadway.Producer.prepare_for_start/2","doc":"Invoked once by Broadway during `Broadway.start_link/2`.\n\nThe goal of this callback is to manipulate the general topology options,\nif necessary at all, and introduce any new child specs that will be\nstarted **before** the producers supervisor in Broadway's supervision tree.\nBroadway's supervision tree is a `rest_for_one` supervisor (see the documentation\nfor `Supervisor`), which means that if the children returned from this callback\ncrash they will bring down the rest of the pipeline before being restarted.\n\nThis callback is guaranteed to be invoked inside the Broadway main process.\n\n`module` is the Broadway module passed as the first argument to\n`Broadway.start_link/2`. `options` is all of Broadway topology options passed\nas the second argument to `Broadway.start_link/2`.\n\nThe return value of this callback is a tuple `{child_specs, options}`. `child_specs`\nis the list of child specs to be started under Broadway's supervision tree.\n`updated_options` is a potentially-updated list of Broadway options\nthat will be used instead of the ones passed to `Broadway.start_link/2`. This can be\nused to modify the characteristics of the Broadway topology to accommodated\nfor the children started here.","ref":"Broadway.Producer.html#c:prepare_for_start/2"},{"type":"callback","title":"Examples - Broadway.Producer.prepare_for_start/2","doc":"defmodule MyProducer do\n      @behaviour Broadway.Producer\n\n      # other callbacks...\n\n      @impl true\n      def prepare_for_start(_module, broadway_options) do\n         children = [\n           {DynamicSupervisor, strategy: :one_for_one, name: MyApp.DynamicSupervisor}\n         ]\n\n         {children, broadway_options}\n      end\n    end","ref":"Broadway.Producer.html#c:prepare_for_start/2-examples"},{"type":"extras","title":"Introduction","doc":"# Introduction\n\n`Broadway` is a library for building concurrent and multi-stage data ingestion and data processing pipelines with Elixir. Broadway pipelines are concurrent and robust, thanks to the Erlang VM and its actors. It features:\n\n  * Back-pressure\n  * Automatic acknowledgements at the end of the pipeline\n  * Batching\n  * Fault tolerance\n  * Graceful shutdown\n  * Built-in testing\n  * Custom failure handling\n  * Ordering and partitioning\n  * Rate-limiting\n  * Metrics","ref":"introduction.html"},{"type":"extras","title":"Official Producers - Introduction","doc":"Currently we officially support four Broadway producers:\n\n  * Amazon SQS: [Source](https://github.com/dashbitco/broadway_sqs) - [Guide](amazon-sqs.md)\n  * Apache Kafka: [Source](https://github.com/dashbitco/broadway_kafka) - [Guide](apache-kafka.md)\n  * Google Cloud Pub/Sub: [Source](https://github.com/dashbitco/broadway_cloud_pub_sub) - [Guide](google-cloud-pubsub.md)\n  * RabbitMQ: [Source](https://github.com/dashbitco/broadway_rabbitmq) - [Guide](rabbitmq.md)\n\nThe guides links above will help you get started with your adapter of choice. For API reference, you can check out the `Broadway` module.","ref":"introduction.html#official-producers"},{"type":"extras","title":"Non-official (Off-Broadway) Producers - Introduction","doc":"For those interested in rolling their own Broadway Producers (which we actively encourage!), we recommend using the `OffBroadway` namespace, mirroring the [Off-Broadway theaters](https://en.wikipedia.org/wiki/Off-Broadway). For example, if you want to publish your own integration with Amazon SQS, you can package it as `off_broadway_sqs`, which uses the `OffBroadway.SQS` namespace.\n\nThe following Off-Broadway libraries are available (feel free to send a PR adding your own in alphabetical order):\n\n  * [off_broadway_amqp10](https://github.com/highmobility/off_broadway_amqp10): [Guide](https://hexdocs.pm/off_broadway_amqp10/)\n  * [off_broadway_elasticsearch](https://github.com/jonlunsford/off_broadway_elasticsearch): [Guide](https://hexdocs.pm/off_broadway_elasticsearch/)\n  * [off_broadway_kafka](https://github.com/bbalser/off_broadway_kafka): [Guide](https://hexdocs.pm/off_broadway_kafka/)\n  * [off_broadway_memory](https://github.com/elliotekj/off_broadway_memory): [Guide](https://hexdocs.pm/off_broadway_memory/)\n  * [off_broadway_redis](https://github.com/amokan/off_broadway_redis): [Guide](https://hexdocs.pm/off_broadway_redis/)\n  * [off_broadway_redis_stream](https://github.com/akash-akya/off_broadway_redis_stream): [Guide](https://hexdocs.pm/off_broadway_redis_stream/)\n  * [off_broadway_splunk](https://github.com/Intility/off_broadway_splunk): [Guide](https://hexdocs.pm/off_broadway_splunk/)","ref":"introduction.html#non-official-off-broadway-producers"},{"type":"extras","title":"Amazon SQS","doc":"# Amazon SQS\n\nAmazon Simple Queue Service (SQS) is a highly scalable distributed message\nqueuing service provided by Amazon.com. AWS SQS offers two types of message\nqueues:\n\n  * Standard\n    * Nearly unlimited throughput\n    * Best-effort ordering\n    * At-least-once delivery\n\n  * FIFO\n    * Limited number of transactions per second (TPS).\n      See [Amazon SQS FIFO](https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/FIFO-queues.html)\n      developer guide for more information on limits.\n    * Order in which messages are sent/received is strictly preserved\n    * Exactly-once delivery\n\nBroadway can work seamlessly with both, Standard and FIFO queues.","ref":"amazon-sqs.html"},{"type":"extras","title":"Getting Started - Amazon SQS","doc":"In order to use Broadway with SQS, we need to:\n\n  1. Create a SQS queue (or use an existing one)\n  1. Configure our Elixir project to use Broadway\n  1. Define your pipeline configuration\n  1. Implement Broadway callbacks\n  1. Run the Broadway pipeline\n  1. Tuning the configuration (Optional)","ref":"amazon-sqs.html#getting-started"},{"type":"extras","title":"Create a SQS queue - Amazon SQS","doc":"Amazon provides a comprehensive [Step-by-step Guide](https://aws.amazon.com/getting-started/tutorials/send-messages-distributed-applications/)\non creating SQS queues. In case you don't have an AWS account and want to\ntest Broadway locally, use can easily install [ElasticMQ](https://github.com/softwaremill/elasticmq),\nwhich is a message queue system that offers a SQS-compatible query interface.","ref":"amazon-sqs.html#create-a-sqs-queue"},{"type":"extras","title":"Configure the project - Amazon SQS","doc":"In this guide we're going to use [BroadwaySQS](https://github.com/dashbitco/broadway_sqs),\nwhich is a Broadway SQS Connector provided by [Dashbit](https://dashbit.co/).","ref":"amazon-sqs.html#configure-the-project"},{"type":"extras","title":"Starting a new project - Amazon SQS","doc":"If you plan to start a new project, just run:\n\n    $ mix new my_app --sup\n\nThe `--sup` flag instructs Elixir to generate an application with a supervision tree.","ref":"amazon-sqs.html#starting-a-new-project"},{"type":"extras","title":"Setting up dependencies - Amazon SQS","doc":"Add `:broadway_sqs` to the list of dependencies in `mix.exs` along the HTTP\nclient of your choice (defaults to `:hackney`):\n\n    def deps do\n      [\n        ...\n        {:broadway_sqs, \"~> 0.7\"},\n        {:hackney, \"~> 1.9\"},\n      ]\n    end\n\nDon't forget to check for the latest version of dependencies.","ref":"amazon-sqs.html#setting-up-dependencies"},{"type":"extras","title":"Define the pipeline configuration - Amazon SQS","doc":"Broadway is a process-based behaviour and to define a Broadway\npipeline, we need to define three functions: `start_link/1`,\n`handle_message/3` and `handle_batch/4`. We will cover `start_link/1`\nin this section and the `handle_` callbacks in the next one.\n\nSimilar to other process-based behaviour, `start_link/1` simply\ndelegates to `Broadway.start_link/2`, which should define the\nproducers, processors, and batchers in the Broadway pipeline.\nAssuming we want to consume messages from a queue called\n`my_queue`, the minimal configuration would be:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module: {BroadwaySQS.Producer,\n                     queue_url: \"https://us-east-2.queue.amazonaws.com/100000000001/my_queue\"}\n          ],\n          processors: [\n            default: []\n          ],\n          batchers: [\n            default: [\n              batch_size: 10,\n              batch_timeout: 2000\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nThe above configuration also assumes that you have the AWS credentials\nset up in your environment, for instance, by having the `AWS_ACCESS_KEY_ID`\nand `AWS_SECRET_ACCESS_KEY` environment variables set. If that's\nnot the case, you will need to pass that information to the client so it\ncan properly connect to the AWS servers. Here is how you can do it:\n\n    ...\n    producer: [\n      module:\n        {BroadwaySQS.Producer,\n         queue_url: \"https://us-east-2.queue.amazonaws.com/100000000001/my_queue\",\n         config: [\n           access_key_id: \"YOUR_AWS_ACCESS_KEY_ID\",\n           secret_access_key: \"YOUR_AWS_SECRET_ACCESS_KEY\"\n         ]}\n    ]\n    ...\n\nFor a full list of options for `BroadwaySQS.Producer`, please see\n[BroadwaySQS](https://hexdocs.pm/broadway_sqs/) documentation.\n\nFor general information about setting up Broadway, see `Broadway`\nmodule docs as well as `Broadway.start_link/2`.\n\n> Note: Even though batching is optional since Broadway v0.2, we recommend that all Amazon SQS\n> pipelines have at least a default batcher. This lets you control the exact batch\n> size and frequency that messages are acknowledged to Amazon SQS, often leading to\n> pipelines that are more cost and time efficient.","ref":"amazon-sqs.html#define-the-pipeline-configuration"},{"type":"extras","title":"Implement Broadway callbacks - Amazon SQS","doc":"In order to process incoming messages, we need to implement the\nrequired callbacks. For the sake of simplicity, we're considering that\nall messages received from the queue are just numbers:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      ...start_link...\n\n      @impl true\n      def handle_message(_, %Message{data: data} = message, _) do\n        message\n        |> Message.update_data(fn data -> data * data end)\n      end\n\n      @impl true\n      def handle_batch(_, messages, _, _) do\n        list = messages |> Enum.map(fn e -> e.data end)\n        IO.inspect(list, label: \"Got batch of finished jobs from processors, sending ACKs to SQS as a batch.\")\n        messages\n      end\n    end\n\nWe are not doing anything fancy here, but it should be enough for our\npurpose. First we update the message's data individually inside\n`handle_message/3` and then we print each batch inside `handle_batch/4`.\n\nFor more information, see `c:Broadway.handle_message/3` and\n`c:Broadway.handle_batch/4`.","ref":"amazon-sqs.html#implement-broadway-callbacks"},{"type":"extras","title":"Run the Broadway pipeline - Amazon SQS","doc":"To run your `Broadway` pipeline, you just need to add as a child in\na supervision tree. Most applications have a supervision tree defined\nat `lib/my_app/application.ex`. You can add Broadway as a child to a\nsupervisor as follows:\n\n    children = [\n      {MyBroadway, []}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n\nNow the Broadway pipeline should be started when your application starts.\nAlso, if your Broadway has any dependency (for example, it needs to talk\nto the database), make sure that Broadway is listed *after* its dependencies\nin the supervision tree.","ref":"amazon-sqs.html#run-the-broadway-pipeline"},{"type":"extras","title":"Tuning the configuration - Amazon SQS","doc":"Some of the configuration options available for Broadway come already with a\n\"reasonable\" default value. However those values might not suit your\nrequirements. Depending on the number of messages you get, how much processing\nthey need and how much IO work is going to take place, you might need completely\ndifferent values to optimize the flow of your pipeline. The `concurrency` option\navailable for every set of producers, processors and batchers, among with\n`max_demand`, `batch_size`, and `batch_timeout` can give you a great deal\nof flexibility.\n\nThe `concurrency` option controls the concurrency level in each layer of\nthe pipeline.\nSee the notes on [`Producer concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-producer-concurrency)\nand [`Batcher concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-batcher-concurrency)\nfor details.\n\nHere's an example on how you could tune them according to\nyour needs.\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            ...\n            concurrency: 10,\n          ],\n          processors: [\n            default: [\n              concurrency: 100,\n              max_demand: 1,\n            ]\n          ],\n          batchers: [\n            default: [\n              batch_size: 10,\n              concurrency: 10,\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nIn order to get a good set of configurations for your pipeline, it's\nimportant to respect the limitations of the servers you're running,\nas well as the limitations of the services you're providing/consuming\ndata to/from. Broadway comes with telemetry, so you can measure your\npipeline and help ensure your changes are effective.","ref":"amazon-sqs.html#tuning-the-configuration"},{"type":"extras","title":"Apache Kafka","doc":"# Apache Kafka\n\nKafka is a distributed streaming platform that has three key capabilities:\n\n  * Publish and subscribe to streams of records\n  * Store streams of records in a fault-tolerant durable way\n  * Process streams of records as they occur","ref":"apache-kafka.html"},{"type":"extras","title":"Getting Started - Apache Kafka","doc":"In order to use Broadway with Kafka, we need to:\n\n  1. Create a stream of records (or use an existing one)\n  1. Configure your Elixir project to use Broadway\n  1. Define your pipeline configuration\n  1. Implement Broadway callbacks\n  1. Run the Broadway pipeline","ref":"apache-kafka.html#getting-started"},{"type":"extras","title":"Create a stream of records (or use an existing one) - Apache Kafka","doc":"In case you don't have Kafka installed yet, please follow the instructions on Kafka's\n[Quickstart](https://kafka.apache.org/quickstart) for a clean installation. After\ninitializing Kafka, you can create a new stream by running:\n\n    $ kafka-topics --create --zookeeper localhost:2181 --partitions 3 --topic test","ref":"apache-kafka.html#create-a-stream-of-records-or-use-an-existing-one"},{"type":"extras","title":"Configure your Elixir project to use Broadway - Apache Kafka","doc":"This guide describes the steps necessary to integrate Broadway with Kafka using\n[BroadwayKafka](https://github.com/dashbitco/broadway_kafka),\nwhich is a Broadway Kafka Connector provided by [Dashbit](https://dashbit.co/).\n\nBroadwayKafka can subscribe to one or more topics and process streams of records\nusing Kafka's [Consumer API](https://kafka.apache.org/documentation.html#consumerapi).\n\nEach GenStage producer initialized by BroadwayKafka will be available as a consumer,\nall registered using the same self-labeled **consumer group**. Each record published to a\ntopic/partition will be delivered to one consumer instance within each consumer group.\n\nBear in mind that a topic/partition can be assigned to any consumer instance that has\nbeen subscribed using the same consumer group, i.e, any Broadway instance or application\nrunning on any machine connected to the Kafka cluster.","ref":"apache-kafka.html#configure-your-elixir-project-to-use-broadway"},{"type":"extras","title":"Starting a new project - Apache Kafka","doc":"Create a new project running:\n\n    $ mix new my_app --sup\n\nThe `--sup` flag instructs Elixir to generate an application with a supervision tree.","ref":"apache-kafka.html#starting-a-new-project"},{"type":"extras","title":"Setting up dependencies - Apache Kafka","doc":"Add `:broadway_kafka` to the list of dependencies in `mix.exs`:\n\n    def deps do\n      [\n        ...\n        {:broadway_kafka, \"~> 0.3\"}\n      ]\n    end\n\nDon't forget to check for the latest version of dependencies.","ref":"apache-kafka.html#setting-up-dependencies"},{"type":"extras","title":"Define the pipeline configuration - Apache Kafka","doc":"Broadway is a process-based behaviour and to define a Broadway pipeline,\nwe need to define three functions: `start_link/1`, `handle_message/3`\nand optionally `handle_batch/4`. We will cover `start_link/1` in this\nsection and the `handle_` callbacks in the next one.\n\nSimilar to other process-based behaviours, `start_link/1` simply\ndelegates to `Broadway.start_link/2`, which should define the\nproducers, processors, and batchers in the Broadway pipeline.\nAssuming we want to consume messages from a topic called\n`test`, one possible configuration would be:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module:\n              {BroadwayKafka.Producer,\n               [\n                 hosts: [localhost: 9092],\n                 group_id: \"group_1\",\n                 topics: [\"test\"]\n               ]},\n            concurrency: 1\n          ],\n          processors: [\n            default: [\n              concurrency: 10\n            ]\n          ],\n          batchers: [\n            default: [\n              batch_size: 100,\n              batch_timeout: 200,\n              concurrency: 10\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\n> **Note**: Pipelines built on top of BroadwayKafka are automatically partitioned.\nSo even though there are multiple processes (stages), these processes will preserve\nKafka's ordering semantics when it comes to topics/partitions. Internally, this is\nachieved by making sure all messages from the same topic/partition will always be\nforwarded to the same processor and batch processor.\n\nFor a full list of options for `BroadwayKafka.Producer`, refer to the\nofficial [BroadwayKafka](https://hexdocs.pm/broadway_kafka/) documentation.\n\nFor general information about setting up Broadway, see `Broadway`\nmodule docs as well as `Broadway.start_link/2`.","ref":"apache-kafka.html#define-the-pipeline-configuration"},{"type":"extras","title":"Implement Broadway callbacks - Apache Kafka","doc":"In order to process incoming messages, we need to implement the\nrequired callbacks. For the sake of simplicity, we're considering that\nall messages received from the topic are just numbers:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      ...start_link...\n\n      @impl true\n      def handle_message(_, message, _) do\n        message\n        |> Message.update_data(fn data -> {data, String.to_integer(data) * 2} end)\n      end\n\n      @impl true\n      def handle_batch(_, messages, _, _) do\n        list = messages |> Enum.map(fn e -> e.data end)\n        IO.inspect(list, label: \"Got batch\")\n        messages\n      end\n    end\n\nWe are not doing anything fancy here, but it should be enough for our\npurpose. First, we update the message's data individually inside\n`handle_message/3` and then we print each batch inside `handle_batch/4`.\n\nFor more information, see `c:Broadway.handle_message/3` and\n`c:Broadway.handle_batch/4`.\n\n> Note: Since Broadway v0.2, batching is optional. In case you don't need to\n> group messages as batches for further processing/publishing, you can remove\n> the `:batchers` configuration along with the `handle_batch/4` callback.","ref":"apache-kafka.html#implement-broadway-callbacks"},{"type":"extras","title":"Run the Broadway pipeline - Apache Kafka","doc":"To run your `Broadway` pipeline, you just need to add as a child in\na supervision tree. Most applications have a supervision tree defined\nat `lib/my_app/application.ex`. You can add Broadway as a child to a\nsupervisor as follows:\n\n    children = [\n      {MyBroadway, []}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n\nNow the Broadway pipeline should be started when your application starts.\nAlso, if your Broadway has any dependency (for example, it needs to talk\nto the database), make sure that Broadway is listed *after* its dependencies\nin the supervision tree.\n\nYou can now test your pipeline by entering an `iex` session:\n\n    $ iex -S mix\n\nIf everything went fine, you should see lots of `info` log messages like this\none coming from the `:brod` supervisors:\n\n    15:14:04.356 [info]  [supervisor: {:local, :brod_sup}, started: [pid: #PID<0.251.0>, id: :test_client, mfargs: {:brod_client, :start_link, [[localhost: 9092], :test_client, []]}, restart_type: {:permanent, 10}, shutdown: 5000, child_type: :worker]]\n\n[Brod](https://github.com/klarna/brod/) is the client that BroadwayKafka uses\nunder the hood to communicate with Kafka.","ref":"apache-kafka.html#run-the-broadway-pipeline"},{"type":"extras","title":"Sending messages to Kafka - Apache Kafka","doc":"Finally, we can send some sample messages to Kafka using using `:brod` with the following snippet:\n\n    topic = \"test\"\n    client_id = :my_client\n    hosts = [localhost: 9092]\n\n    :ok = :brod.start_client(hosts, client_id, _client_config=[])\n    :ok = :brod.start_producer(client_id, topic, _producer_config = [])\n\n    Enum.each(1..1000, fn i ->\n      partition = rem(i, 3)\n      :ok = :brod.produce_sync(client_id, topic, partition, _key=\"\", \"#{i}\")\n    end)\n\nYou should see the output showing the generated batches:\n\n    Got batch: [\n      {\"2\", 4},\n      {\"5\", 10},\n      {\"8\", 16},\n      {\"11\", 22},\n      {\"14\", 28},\n      ...\n    ]\n    Got batch: [\n      {\"3\", 6},\n      {\"6\", 12},\n      {\"9\", 18},\n      {\"12\", 24},\n      {\"15\", 30},\n      ...\n    ]","ref":"apache-kafka.html#sending-messages-to-kafka"},{"type":"extras","title":"Tuning the configuration - Apache Kafka","doc":"Some of the configuration options available for Broadway come already with a\n\"reasonable\" default value. However, those values might not suit your\nrequirements. Depending on the number of records you get, how much processing\nthey need and how much IO work is going to take place, you might need completely\ndifferent values to optimize the flow of your pipeline. The `concurrency` option\navailable for every set of producers, processors and batchers, along with\n`batch_size` and `batch_timeout` can give you a great deal of flexibility.\nSee the notes on [`Producer concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-producer-concurrency)\nand [`Batcher concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-batcher-concurrency)\nfor details.\n\nBy setting the `concurrency` option, you define the number of concurrent processes\nthat will be started by Broadway, allowing you to have full control over the\nconcurrency level in each layer of the pipeline. Keep in mind that since the\nconcurrency model provided by **Kafka** is based on **partitioning**, in order to take\nfull advantage of this model, you need to set the `concurrency` option for\nyour processors and batchers accordingly. Having less concurrency than topic/partitions\nassigned will result in individual processors handling more than one partition,\ndecreasing the overall level of concurrency. Therefore, if you want to always be able\nto process messages at maximum concurrency (assuming you have enough resources to do it),\nyou should increase the concurrency up front to make sure you have enough processors to\nhandle the extra records received from new partitions assigned.\n\n> **Note**: Even if you don't plan to add more partitions to a Kafka topic, your pipeline\ncan still receive more assignments than planned. For instance, if another consumer crashes,\nthe server will reassign all its topic/partition to other available consumers, including\nany Broadway producer subscribed to the same topic.\n\nThere are other options that you may want to take a closer look when tuning your configuration.\nThe `:max_bytes` option, for instance, belongs to the `:fetch_config` group and defines the\nmaximum amount of data to be fetched at a time from a single partition. The default is\n1048576 (1 MiB). Setting greater values can improve throughput at the cost of more\nmemory consumption. For more information and other fetch options, please refer to the\n\"Fetch config options\" in the official [BroadwayKafka](https://hexdocs.pm/broadway_kafka/)\ndocumentation.\n\nOther two important options are `:offset_commit_interval_seconds` and `:offset_commit_on_ack`.\nBoth belong to the main configuration and they can make a huge impact on performance.\n\nThe `:offset_commit_interval_seconds` defines the time interval between two\nOffsetCommitRequest messages. The default is 5s.\n\nThe `:offset_commit_on_ack`, when set to `true`, tells Broadway to send an\nOffsetCommitRequest immediately after each acknowledgement, bypassing any\ninterval defined in `:offset_commit_interval_seconds`. Setting this option to\n`false` can increase performance since any commit requests will start respecting\nthe `:offset_commit_interval_seconds` option. This will usually result in fewer\nrequests to be sent to the server. However, setting long commit intervals might\nlead to a large number of duplicated records to be processed after a server\nrestart or connection loss. Since it is always possible that duplicate messages\nwill be received by consumers, make sure your logic is idempotent when consuming\nrecords to avoid inconsistencies. Also, bear in mind that the negative\nperformance impact might be insignificant if you're using batchers since only\none commit request will be performed per batch. As a basic rule, always take\ninto account the values of `batch_size` and `batch_timeout` whenever you're\ntuning `:offset_commit_interval_seconds` and `:offset_commit_on_ack`.","ref":"apache-kafka.html#tuning-the-configuration"},{"type":"extras","title":"Handling failed messages - Apache Kafka","doc":"`broadway_kafka` never stops the flow of the stream, i.e. it will **always ack** the messages\neven when they fail. Unlike queue-based connectors, where you can mark a single message as failed.\nIn Kafka that's not possible due to its single offset per topic/partition ack strategy. If you\nwant to reprocess failed messages, you need to roll your own strategy. A possible way to do that\nis to implement `handle_failed/2` and send failed messages to a separated stream or queue for\nlater processing.","ref":"apache-kafka.html#handling-failed-messages"},{"type":"extras","title":"Google Cloud Pub/Sub","doc":"# Google Cloud Pub/Sub\n\nCloud Pub/Sub is a fully-managed real-time messaging service provided by Google.","ref":"google-cloud-pubsub.html"},{"type":"extras","title":"Getting Started - Google Cloud Pub/Sub","doc":"In order to use Broadway with Cloud Pub/Sub you need to:\n\n  1. Setup a Cloud Pub/Sub project\n  1. Configure your Elixir project to use Broadway\n  1. Define your pipeline configuration\n  1. Implement Broadway callbacks\n  1. Run the Broadway pipeline\n  1. Tune the configuration (Optional)\n\nIf you are just getting familiar with Google Pub/Sub, refer to [the documentation](https://cloud.google.com/pubsub/docs/)\nto get started. Instead of testing against a live environment, you may also consider using the\n[emulator](https://cloud.google.com/pubsub/docs/emulator) to simulate integrating with Cloud\nPub/Sub.\n\nIf you have an existing project, topic, subscription, and credentials, you can skip [step\n1](#setup-cloud-pub-sub-project) and jump to [Configure the project](#configure-the-project)\nsection.","ref":"google-cloud-pubsub.html#getting-started"},{"type":"extras","title":"Setup Cloud Pub/Sub project - Google Cloud Pub/Sub","doc":"In this tutorial we'll use the [`gcloud`](https://cloud.google.com/sdk/gcloud/) command-line tool\nto set everything up in Google Cloud. Alternatively, you can roughly follow this guide by using\n[Cloud Console](https://console.cloud.google.com).\n\nTo install `gcloud` follow the [documentation](https://cloud.google.com/sdk/gcloud/). If you are\non macOS you may consider installing it with Homebrew:\n\n    $ brew install --cask google-cloud-sdk\n\nNow, authenticate the CLI:\n\n    $ gcloud auth login\n\nThen, create a new project:\n\n    $ gcloud projects create test-pubsub\n\nA new topic:\n\n    $ gcloud pubsub topics create test-topic --project test-pubsub\n    Created topic [projects/test-pubsub/topics/test-topic].\n\n> Note: If you run this command immediately after creating a new Google Cloud project, you may receive an error indicating that your project's organization policy is still being provisioned. Just wait a couple minutes and try again.\n\nAnd a new subscription:\n\n    $ gcloud pubsub subscriptions create test-subscription --project test-pubsub --topic test-topic\n    Created subscription [projects/test-pubsub/subscriptions/test-subscription].\n\nWe also need a [service account](https://cloud.google.com/iam/docs/service-accounts), an IAM\npolicy, as well as API credentials in order to programmatically work with the service. First, let's\ncreate the service account:\n\n    $ gcloud iam service-accounts create test-account --project test-pubsub\n    Created service account [test-account].\n\nThen the policy. For simplicity we add the general role `roles/editor`, but make sure to\nexamine the [available roles](https://cloud.google.com/iam/docs/understanding-roles#pubsub-roles)\nand choose the one that best suits your use case:\n\n    $ gcloud projects add-iam-policy-binding test-pubsub \\\n        --member serviceAccount:test-account@test-pubsub.iam.gserviceaccount.com \\\n        --role roles/editor\n    Updated IAM policy for project [test-pubsub].\n    (...)\n\nAnd now the credentials:\n\n    $ gcloud iam service-accounts keys create credentials.json --iam-account=test-account@test-pubsub.iam.gserviceaccount.com\n    created key [xxx] of type [json] as [key] for [test-account@test-pubsub.iam.gserviceaccount.com]\n\nThis command generated a `credentials.json` file which will be useful later. Note, the IAM account\npattern is ` @ .iam.gserviceaccount.com`. Run `gcloud iam service-accounts list --project test-pubsub`\nto see all service accounts associated with the given project.\n\nFinally, we need to enable Pub/Sub for our project:\n\n    $ gcloud services enable pubsub --project test-pubsub\n    Operation \"operations/xxx\" finished successfully.","ref":"google-cloud-pubsub.html#setup-cloud-pub-sub-project"},{"type":"extras","title":"Configure the project - Google Cloud Pub/Sub","doc":"In this guide we're going to use [BroadwayCloudPubSub](https://github.com/dashbitco/broadway_cloud_pub_sub),\nwhich is a Broadway Cloud Pub/Sub Connector provided by [Dashbit](https://dashbit.co/).","ref":"google-cloud-pubsub.html#configure-the-project"},{"type":"extras","title":"Starting a new project - Google Cloud Pub/Sub","doc":"If you plan to start a new project, just run:\n\n    $ mix new my_app --sup\n\nThe `--sup` flag instructs Elixir to generate an application with a supervision tree.","ref":"google-cloud-pubsub.html#starting-a-new-project"},{"type":"extras","title":"Setting up dependencies - Google Cloud Pub/Sub","doc":"Add `:broadway_cloud_pub_sub` to the list of dependencies in `mix.exs`, along with the Google\nCloud authentication library of your choice (defaults to `:goth`):\n\n    defp deps() do\n      [\n        ...\n        {:broadway_cloud_pub_sub, \"~> 0.7\"},\n        {:goth, \"~> 1.0\"}\n      ]\n    end\n\nDon't forget to check for the latest version of dependencies.","ref":"google-cloud-pubsub.html#setting-up-dependencies"},{"type":"extras","title":"Define the pipeline configuration - Google Cloud Pub/Sub","doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three\nfunctions: `start_link/1`, `handle_message/3` and `handle_batch/4`. We will cover `start_link/1`\nin this section and the `handle_` callbacks in the next one.\n\nSimilar to other process-based behaviour, `start_link/1` simply delegates to\n`Broadway.start_link/2`, which should define the producers, processors, and batchers in the\nBroadway pipeline. Assuming we want to consume messages from the `test-subscription`, the minimal\nconfiguration would be:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module:\n              {BroadwayCloudPubSub.Producer,\n               subscription: \"projects/test-pubsub/subscriptions/test-subscription\"}\n          ],\n          processors: [\n            default: []\n          ],\n          batchers: [\n            default: [\n              batch_size: 10,\n              batch_timeout: 2_000\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nFor a full list of options for `BroadwayCloudPubSub.Producer`, please see [the\ndocumentation](https://hexdocs.pm/broadway_cloud_pub_sub).\n\nFor general information about setting up Broadway, see `Broadway` module docs as well as\n`Broadway.start_link/2`.\n\n> Note: Even though batching is optional since Broadway v0.2, we recommend all Cloud Pub/Sub\n> pipelines to have at least a default batcher, as that allows you to control the exact batch\n> size and frequency that messages are acknowledged to Cloud Pub/Sub, which often leads to\n> pipelines that are more cost and time efficient.","ref":"google-cloud-pubsub.html#define-the-pipeline-configuration"},{"type":"extras","title":"Implement Broadway callbacks - Google Cloud Pub/Sub","doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake\nof simplicity, we're considering that all messages received from the queue are strings and our\nprocessor calls `String.upcase/1` on them:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      ...start_link...\n\n      def handle_message(_, %Message{data: data} = message, _) do\n        message\n        |> Message.update_data(fn data -> String.upcase(data) end)\n      end\n\n      def handle_batch(_, messages, _, _) do\n        list = messages |> Enum.map(fn e -> e.data end)\n        IO.inspect(list, label: \"Got batch of finished jobs from processors, sending ACKs to Pub/Sub as a batch.\")\n        messages\n      end\n    end\n\nWe are not doing anything fancy here, but it should be enough for our purpose. First we update the\nmessage's data individually inside `handle_message/3` and then we print each batch inside\n`handle_batch/4`.\n\nFor more information, see `c:Broadway.handle_message/3` and `c:Broadway.handle_batch/4`.","ref":"google-cloud-pubsub.html#implement-broadway-callbacks"},{"type":"extras","title":"Run the Broadway pipeline - Google Cloud Pub/Sub","doc":"To run your `Broadway` pipeline, you need to add it as a child in a supervision tree. Most\napplications have a supervision tree defined at `lib/my_app/application.ex`. You can add Broadway\nas a child to a supervisor as follows:\n\n    children = [\n      {MyBroadway, []}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n\nThe final step is to configure credentials. You can set the following environment variable:\n\n    export GOOGLE_APPLICATION_CREDENTIALS=/path/to/credentials.json\n\nSee [Goth](https://github.com/peburrows/goth) documentation for alternative ways of authenticating\nwith the API.\n\nNow the Broadway pipeline should be started when your application starts. Also, if your Broadway\npipeline has any dependency (for example, it needs to talk to the database), make sure that\nit is listed *after* its dependencies in the supervision tree.\n\nIf you followed the previous section about setting the project with `gcloud`, you can now test the\nthe pipeline. In one terminal tab start the application:\n\n    $ iex -S mix\n\nAnd in another tab, send a couple of test messages to Pub/Sub:\n\n    $ gcloud pubsub topics publish  projects/test-pubsub/topics/test-topic --message \"test 1\"\n    messageIds:\n    - '651428033718119'\n\n    gcloud pubsub topics publish  projects/test-pubsub/topics/test-topic --message \"test 2\"\n    messageIds:\n    - '651427034966696'\n\nNow, In the first tab, you should see output similar to:\n\n```\nGot batch of finished jobs from processors, sending ACKs to Pub/Sub as a batch.: [\"TEST 1\", \"TEST 2\"]\n```","ref":"google-cloud-pubsub.html#run-the-broadway-pipeline"},{"type":"extras","title":"Tuning the configuration - Google Cloud Pub/Sub","doc":"Some of the configuration options available for Broadway come already with a\n\"reasonable\" default value. However those values might not suit your\nrequirements. Depending on the number of messages you get, how much processing\nthey need and how much IO work is going to take place, you might need completely\ndifferent values to optimize the flow of your pipeline. The `concurrency` option\navailable for every set of producers, processors and batchers, among with\n`max_demand`, `batch_size`, and `batch_timeout` can give you a great deal\nof flexibility.\n\nThe `concurrency` option controls the concurrency level in each layer of\nthe pipeline.\nSee the notes on [`Producer concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-producer-concurrency)\nand [`Batcher concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-batcher-concurrency)\nfor details.\n\nHere's an example on how you could tune them according to\nyour needs.\n\n    defmodule MyBroadway do\n      use Broadway\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            ...\n            concurrency: 10,\n          ],\n          processors: [\n            default: [\n              concurrency: 100,\n              max_demand: 1,\n            ]\n          ],\n          batchers: [\n            default: [\n              batch_size: 10,\n              concurrency: 10,\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nIn order to get a good set of configurations for your pipeline, it's\nimportant to respect the limitations of the servers you're running,\nas well as the limitations of the services you're providing/consuming\ndata to/from. Broadway comes with telemetry, so you can measure your\npipeline and help ensure your changes are effective.","ref":"google-cloud-pubsub.html#tuning-the-configuration"},{"type":"extras","title":"RabbitMQ","doc":"# RabbitMQ\n\nRabbitMQ is an open source message broker designed to be highly scalable and\ndistributed. It supports multiple protocols including the Advanced Message\nQueuing Protocol (AMQP).","ref":"rabbitmq.html"},{"type":"extras","title":"Getting Started - RabbitMQ","doc":"In order to use Broadway with RabbitMQ, we need to:\n\n  1. [Create a queue](#create-a-queue) (or use an existing one)\n  1. [Configure our Elixir project to use Broadway](#configure-the-project)\n  1. [Define your pipeline configuration](#define-the-pipeline-configuration)\n  1. [Implement Broadway callbacks](#implement-broadway-callbacks)\n  1. [Run the Broadway pipeline](#run-the-broadway-pipeline)\n  1. [Tuning the configuration](#tuning-the-configuration) (Optional)\n\nIn case you want to work with an existing queue, you can skip [step 1](#create-a-queue)\nand jump to [Configure the project](#configure-the-project).\n\n> Note: `BroadwayRabbitMQ` does not automatically create any queue. If you\nconfigure a pipeline with a non-existent queue, the producers will crash,\nbringing down the pipeline.","ref":"rabbitmq.html#getting-started"},{"type":"extras","title":"Create a queue - RabbitMQ","doc":"RabbitMQ runs on many operating systems. Please see\n[Downloading and Installing RabbitMQ](https://www.rabbitmq.com/download.html) for\nfurther information. Also, make sure you have the\n[Management](https://www.rabbitmq.com/management.html) plugin enabled, which ships\nwith the command line tool, `rabbitmqadmin`.\n\nAfter successfully installing RabbitMQ, you can declare a new queue with the\nfollowing command:\n\n    $ rabbitmqadmin declare queue name=my_queue durable=true\n\nYou can list all declared queues to see our the one we've just created:\n\n    $ rabbitmqctl list_queues\n    Timeout: 60.0 seconds ...\n    Listing queues for vhost / ...\n    name      messages\n    my_queue  0","ref":"rabbitmq.html#create-a-queue"},{"type":"extras","title":"Configure the project - RabbitMQ","doc":"In this guide, we're going to use [BroadwayRabbitMQ](https://github.com/dashbitco/broadway_rabbitmq),\nwhich is a Broadway RabbitMQ Connector provided by [Dashbit](https://dashbit.co/).","ref":"rabbitmq.html#configure-the-project"},{"type":"extras","title":"Starting a new project - RabbitMQ","doc":"If you're creating a new project, run:\n\n    $ mix new my_app --sup\n\nThe `--sup` flag instructs Elixir to generate an application with a supervision tree.","ref":"rabbitmq.html#starting-a-new-project"},{"type":"extras","title":"Setting up dependencies - RabbitMQ","doc":"Add `:broadway_rabbitmq` to the list of dependencies in `mix.exs`:\n\n    def deps do\n      [\n        ...\n        {:broadway_rabbitmq, \"~> 0.7\"},\n      ]\n    end\n\nDon't forget to check for the latest version of dependencies.","ref":"rabbitmq.html#setting-up-dependencies"},{"type":"extras","title":"Define the pipeline configuration - RabbitMQ","doc":"Broadway is a process-based behaviour and to define a Broadway pipeline,\nwe need to define three functions: `start_link/1`, `handle_message/3`\nand optionally `handle_batch/4`. We will cover `start_link/1` in this\nsection and the `handle_` callbacks in the next one.\n\nSimilar to other process-based behaviours, `start_link/1` simply\ndelegates to `Broadway.start_link/2`, which should define the\nproducers, processors, and batchers in the Broadway pipeline.\nAssuming we want to consume messages from a queue called\n`my_queue`, one possible configuration would be:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: MyBroadway,\n          producer: [\n            module: {BroadwayRabbitMQ.Producer,\n              queue: \"my_queue\",\n              qos: [\n                prefetch_count: 50,\n              ]\n            },\n            concurrency: 1\n          ],\n          processors: [\n            default: [\n              concurrency: 50\n            ]\n          ],\n          batchers: [\n            default: [\n              batch_size: 10,\n              batch_timeout: 1500,\n              concurrency: 5\n            ]\n          ]\n        )\n      end\n\n      ...callbacks...\n    end\n\nIf you're consuming data from an existing broker that requires authorization,\nyou'll need to provide your credentials using the `connection` option:\n\n    ...\n    producer: [\n      module: {BroadwayRabbitMQ.Producer,\n        queue: \"my_queue\",\n        connection: [\n          username: \"user\",\n          password: \"password\",\n        ]\n        ...\n      }\n    ]\n    ...\n\nFor the full list of `connection` options, please see\n[`AMQP.Connection.open/1`](https://hexdocs.pm/amqp/1.1.1/AMQP.Connection.html#open/1)\n\nFor general information about setting up Broadway, see `Broadway`\nmodule docs as well as `Broadway.start_link/2`.","ref":"rabbitmq.html#define-the-pipeline-configuration"},{"type":"extras","title":"Implement Broadway callbacks - RabbitMQ","doc":"In order to process incoming messages, we need to implement the\nrequired callbacks. For the sake of simplicity, we're considering that\nall messages received from the queue are just numbers:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      ...start_link...\n\n      @impl true\n      def handle_message(_, message, _) do\n        message\n        |> Message.update_data(fn data -> {data, String.to_integer(data) * 2} end)\n      end\n\n      @impl true\n      def handle_batch(_, messages, _, _) do\n        list = messages |> Enum.map(fn e -> e.data end)\n        IO.inspect(list, label: \"Got batch\")\n        messages\n      end\n    end\n\nWe are not doing anything fancy here, but it should be enough for our\npurpose. First, we update the message's data individually inside\n`handle_message/3` and then we print each batch inside `handle_batch/4`.\n\nFor more information, see `c:Broadway.handle_message/3` and\n`c:Broadway.handle_batch/4`.\n\n> Note: Since Broadway v0.2, batching is optional. In case you don't need to\n> group messages as batches for further processing/publishing, you can remove\n> the `:batchers` configuration along with the `handle_batch/4` callback. This\n> is perfectly fine for RabbitMQ, where messages are acknowledged individually\n> and never as a batch.","ref":"rabbitmq.html#implement-broadway-callbacks"},{"type":"extras","title":"Run the Broadway pipeline - RabbitMQ","doc":"To run your `Broadway` pipeline, you just need to add as a child in\na supervision tree. Most applications have a supervision tree defined\nat `lib/my_app/application.ex`. You can add Broadway as a child to a\nsupervisor as follows:\n\n    children = [\n      {MyBroadway, []}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n\nNow the Broadway pipeline should be started when your application starts.\nAlso, if your Broadway has any dependency (for example, it needs to talk\nto the database), make sure that Broadway is listed *after* its dependencies\nin the supervision tree.\n\nYou can now test your pipeline by entering an `iex` session:\n\n    $ iex -S mix\n\nIf everything went fine, you should see lots of `info` log messages from the `amqp`\nsupervisors. If you think that's too verbose and want to do something\nabout it, please take a look at the _\"Log related to amqp supervisors are too verbose\"_\nsubsection in the `amqp`'s  [Troubleshooting](https://hexdocs.pm/amqp/readme.html#troubleshooting)\ndocumentation.\n\nFinally, let's generate some sample messages to be consumed by Broadway with the\nfollowing code:\n\n    {:ok, connection} = AMQP.Connection.open\n    {:ok, channel} = AMQP.Channel.open(connection)\n    AMQP.Queue.declare(channel, \"my_queue\", durable: true)\n\n    Enum.each(1..5000, fn i ->\n      AMQP.Basic.publish(channel, \"\", \"my_queue\", \"#{i}\")\n    end)\n    AMQP.Connection.close(connection)\n\nYou should see the output showing the generated batches:\n\n    Got batch: [\n      {\"7\", 14},\n      {\"5\", 10},\n      {\"8\", 16},\n      {\"98\", 196},\n      {\"6\", 12},\n      {\"97\", 194},\n      {\"9\", 18},\n      {\"99\", 198},\n      {\"10\", 20},\n      {\"100\", 200}\n    ]\n    Got batch: [\n      {\"29\", 58},\n      {\"32\", 64},\n      ...\n    ]","ref":"rabbitmq.html#run-the-broadway-pipeline"},{"type":"extras","title":"Tuning the configuration - RabbitMQ","doc":"Some of the configuration options available for Broadway come already with a\n\"reasonable\" default value. However, those values might not suit your\nrequirements. Depending on the number of messages you get, how much processing\nthey need and how much IO work is going to take place, you might need completely\ndifferent values to optimize the flow of your pipeline. The `concurrency` option\navailable for every set of producers, processors and batchers, among with\n`max_demand`, `batch_size`, and `batch_timeout` can give you a great deal\nof flexibility. The `concurrency` option controls the concurrency level in\neach layer of the pipeline.\nSee the notes on [`Producer concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-producer-concurrency)\nand [`Batcher concurrency`](https://hexdocs.pm/broadway/Broadway.html#module-batcher-concurrency)\nfor details.\n\nAnother important option to take into account is the `:prefetch_count`.\nRabbitMQ will continually push new messages to Broadway as it receives them.\nThe `:prefetch_count` setting provides back-pressure by instructing RabbitMQ to [limit the number of unacknowledged messages a consumer will have at a given moment](https://www.rabbitmq.com/consumer-prefetch.html).\nSee the [\"Back-pressure and :prefetch_count\"](https://hexdocs.pm/broadway_rabbitmq/BroadwayRabbitMQ.Producer.html#module-back-pressure-and-prefetch_count)\nsection of the `BroadwayRabbitMQ` documentation for details.\n\nIn order to get a good set of configurations for your pipeline, it's\nimportant to respect the limitations of the servers you're running,\nas well as the limitations of the services you're providing/consuming\ndata to/from. Broadway comes with telemetry, so you can measure your\npipeline and help ensure your changes are effective.","ref":"rabbitmq.html#tuning-the-configuration"},{"type":"extras","title":"Custom Producers","doc":"# Custom Producers\n\nIf you want to use Broadway but there is no existing Broadway producer\nfor the technology of your choice, you can integrate any existing GenStage\nproducer into the pipeline with relative ease.","ref":"custom-producers.html"},{"type":"extras","title":"Example - Custom Producers","doc":"In general, producers must generate `%Broadway.Message{}` structs in order\nto be processed by Broadway. In case you need to use an existing GenStage\nproducer and you don't want to change its original implementation,\nyou'll have to set the producer's `:transformer` option to translate the\ngenerated events into Broadway messages.\n\nIn the following example the producer is a regular GenStage, i.e., it\nproduces plain events that cannot be processed by Broadway directly:\n\n    defmodule Counter do\n      use GenStage\n\n      def start_link(number) do\n        GenStage.start_link(Counter, number)\n      end\n\n      def init(counter) do\n        {:producer, counter}\n      end\n\n      def handle_demand(demand, counter) when demand > 0 do\n        events = Enum.to_list(counter..counter+demand-1)\n        {:noreply, events, counter + demand}\n      end\n    end\n\nBy using a transformer, you can tell Broadway to transform all events\ngenerated by the producer into proper Broadway messages:\n\n    defmodule MyBroadway do\n      use Broadway\n\n      alias Broadway.Message\n\n      def start_link(_opts) do\n        Broadway.start_link(__MODULE__,\n          name: __MODULE__,\n          producer: [\n            module: {Counter, 1},\n            transformer: {__MODULE__, :transform, []}\n          ],\n          processors: [\n            default: [concurrency: 10]\n          ],\n          batchers: [\n            default: [concurrency: 2, batch_size: 5],\n          ]\n        )\n      end\n\n      ...callbacks...\n\n      def transform(event, _opts) do\n        %Message{\n          data: event,\n          acknowledger: {__MODULE__, :ack_id, :ack_data}\n        }\n      end\n\n      def ack(:ack_id, successful, failed) do\n        # Write ack code here\n      end\n    end\n\nNotice that you need to pass two options to the producer:\n\n  * `:module` - a tuple representing the GenStage producer as `{mod, arg}`.\n    Where `mod` is module that implements the GenStage behaviour and `arg`\n    the argument that will be given to the `init` callback of the GenStage.\n    It is very important to note that Broadway **will not call** the\n    `child_spec/1` or `start_link/1` function of the producer. That's\n    because Broadway wraps the producer to augment it with extra features.\n\n  * `:transformer` - a module-function-args tuple that will be invoked\n    inside the producer, for every producer message, that should create\n    a `Broadway.Message` struct with the `data` and `acknowledger` fields.\n\nSee the `Broadway.Acknowledger` module for more information on defining\nand setting up acknowledgements.","ref":"custom-producers.html#example"},{"type":"extras","title":"Architecture","doc":"# Architecture\n\nBroadway's architecture is built on top of GenStage. That means we structure\nour processing units as independent stages that are responsible for one\nindividual task in the pipeline. By implementing the `Broadway` behaviour,\nwe define a `GenServer` process that wraps a `Supervisor` to manage and\nown our pipeline.","ref":"architecture.html"},{"type":"extras","title":"The pipeline model - Architecture","doc":"```asciidoc\n                                       [producers]   <- pulls data from SQS, RabbitMQ, etc.\n                                            |\n                                            |   (demand dispatcher)\n                                            |\n   handle_message/3 and  ---------->   [processors]\n   prepare_messages/2 run here             / \\\n                                          /   \\   (partition dispatcher)\n                                         /     \\\n                                   [batcher]   [batcher]   <- one for each batcher key\n                                       |           |\n                                       |           |   (demand dispatcher)\n                                       |           |\nhandle_batch/4 runs here -> [batch processor][batch processor]\n```","ref":"architecture.html#the-pipeline-model"},{"type":"extras","title":"Internal stages - Architecture","doc":"* `Broadway.Producer` - A wrapper around the actual producer defined by\n    the user. It serves as the source of the pipeline.\n  * `Broadway.Processor` - This is where messages are processed, e.g. do\n    calculations, convert data into a custom json format etc. Here is where\n    the code from `handle_message/3` runs.\n  * `Broadway.Batcher` - Creates batches of messages based on the\n    batcher's key. One batcher for each key will be created.\n  * `Broadway.BatchProcessor` - This is where the code from `handle_batch/4` runs.","ref":"architecture.html#internal-stages"},{"type":"extras","title":"The supervision tree - Architecture","doc":"Broadway was designed to always go back to a working state in case\nof failures thanks to the use of supervisors. Our supervision tree\nis designed as follows:\n\n```asciidoc\n                                   [Broadway GenServer]\n                                            |\n                                            |\n                                            |\n                              [Broadway Pipeline Supervisor]\n                             /    /   (:rest_for_one)   \\    \\\n                           /     |                       |      \\\n                         /       |                       |         \\\n                       /         |                       |            \\\n                     /           |                       |               \\\n                   /             |                       |                  \\\n  [ProducerSupervisor]  [ProcessorSupervisor]   [BatchersSupervisor]    [Terminator]\n    (:one_for_one)        (:one_for_all)           (:one_for_one)\n         / \\                    / \\                /            \\\n        /   \\                  /   \\              /              \\\n       /     \\                /     \\            /                \\\n      /       \\              /       \\          /                  \\\n[Producer_1]  ...    [Processor_1]  ...  [BatcherSupervisor_1]     ...\n                                            (:rest_for_one)\n                                            /             \\\n                                           /               \\\n                                          /                 \\\n                                     [Batcher]   [BatchProcessorSupervisor]\n                                                       (:one_for_all)\n                                                        /           \\\n                                                       /             \\\n                                                      /               \\\n                                            [BatchProcessor_1]        ...\n```\n\n\nBoth `ProcessorSupervisor` and `BatchProcessorSupervisor` are set with\n`max_restarts` to 0. The idea is that if any process fails, we want\nto restart the rest of the tree. Since Broadway callbacks are\nstateless, we can handle errors and provide reports without crashing\nprocesses. This means that the supervision tree will only shutdown\nin case of unforeseen errors in Broadway's implementation.\n\nThe only exception are the producers, which contain external code\nand are expected to fail. If a producer crashes, it will be restarted\nby its supervisor without cascading failures until its max restarts\nis reached. Broadway automatically handles those failures by making\nprocessors automatically resubscribe to producers in case of crashes.","ref":"architecture.html#the-supervision-tree"},{"type":"extras","title":"Graceful shutdowns - Architecture","doc":"The cascading failures aspect also provides safe semantics for graceful\nshutdown. We know that either all processes are running OR they are all\nbeing shutdown. Therefore, to gracefully shutdown the supervision tree,\na terminator process is activated, which starts the following steps:\n\n  1. It notifies the first layer of processors that they should not\n     resubscribe to producers once they exit\n\n  2. It tells all producers to no longer accept demand, flush all\n     current events, and then shutdown\n\n  3. It then monitors and waits for a confirmation message from batch\n     processors. At this point, the terminator is effectively blocking\n     the supervisor until all events have been processed\n\nThis triggers a cascade effect where processors notice all of its producers\nhave been cancelled, causing them to flush their own events and cancels the\nstages downstream, and so on and so on. This happens until batch processors\nnotice all of their producers have been cancelled, effectively notifying the\nterminator to shutdown, allowing the outer most supervisor to go on and fully\nterminate all stages, which at this point have flushed all events.","ref":"architecture.html#graceful-shutdowns"}],"proglang":"elixir","producer":{"name":"ex_doc","version":"0.37.1"},"content_type":"text/markdown"}